# Copyright (c) 2025 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)

import os
import pickle
import pytest

import numpy as np

import gen_tso.pandeia_io as jwst
from gen_tso.utils import ROOT

os.chdir(ROOT+'../tests')
# See tests/mocks/make_mocks.py for mocked data setup.


@pytest.mark.skip(reason='TBD')
def test_read_noise_variance():
    pass


@pytest.mark.parametrize('nint', [1, 10, 100])
def test_exposure_time_nircam(nint):
    inst = 'nircam'
    subarray = 'subgrism64'
    readout = 'rapid'
    ngroup = 90
    exp_time = jwst.exposure_time(inst, subarray, readout, ngroup, nint)
    if nint == 1:
        expected_exposure = 31.00063
    elif nint == 10:
        expected_exposure = 310.0063
    elif nint == 100:
        expected_exposure = 3100.063
    np.testing.assert_allclose(exp_time, expected_exposure)


@pytest.mark.parametrize('nint', [1, 10, 100])
def test_exposure_time_miri_fastr1(nint):
    inst = 'miri'
    subarray = 'slitlessprism'
    readout = 'fastr1'
    ngroup = 30
    exp_time = jwst.exposure_time(inst, subarray, readout, ngroup, nint)
    if nint == 1:
        expected_exposure = 4.7712
    elif nint == 10:
        expected_exposure = 49.14336
    elif nint == 100:
        expected_exposure = 492.86496
    np.testing.assert_allclose(exp_time, expected_exposure)


@pytest.mark.parametrize('nint', [1, 10, 100])
def test_exposure_time_miri_slowr1(nint):
    inst = 'miri'
    subarray = 'full'
    readout = 'slowr1'
    ngroup = 30
    exp_time = jwst.exposure_time(inst, subarray, readout, ngroup, nint)
    if nint == 1:
        expected_exposure = 716.6976
    elif nint == 10:
        expected_exposure = 7381.98528
    elif nint == 100:
        expected_exposure = 74034.86208
    np.testing.assert_allclose(exp_time, expected_exposure)


def test_exposure_time_bad_subarray():
    inst = 'nircam'
    subarray = 'nope'
    readout = 'rapid'
    ngroup = 90
    nint = 1
    exp_time = jwst.exposure_time(inst, subarray, readout, ngroup, nint)
    assert exp_time == 0.0


def test_exposure_time_bad_readout():
    inst = 'nircam'
    subarray = 'subgrism64'
    readout = 'nope'
    ngroup = 90
    nint = 1
    exp_time = jwst.exposure_time(inst, subarray, readout, ngroup, nint)
    assert exp_time == 0.0


def test_bin_search_exposure_time_nircam():
    inst = 'nircam'
    subarray = 'subgrism64'
    readout = 'rapid'
    ngroup = 90
    obs_time = 6.0
    nint, exp_time = jwst.bin_search_exposure_time(
        inst, subarray, readout, ngroup, obs_time,
    )
    assert nint == 697
    np.testing.assert_allclose(exp_time, 21607.43911)


def test_bin_search_exposure_time_miri():
    inst = 'miri'
    subarray = 'slitlessprism'
    readout = 'fastr1'
    ngroup = 30
    obs_time = 6.0
    nint, exp_time = jwst.bin_search_exposure_time(
        inst, subarray, readout, ngroup, obs_time,
    )
    assert nint == 4382
    np.testing.assert_allclose(exp_time, 21604.15264)


def test_bin_search_exposure_time_bad_subarray():
    inst = 'nircam'
    subarray = 'nope'
    readout = 'rapid'
    ngroup = 90
    obs_time = 6.0
    nint, exp_time = jwst.bin_search_exposure_time(
        inst, subarray, readout, ngroup, obs_time,
    )
    assert exp_time == 0.0


def test_bin_search_exposure_time_bad_readout():
    inst = 'nircam'
    subarray = 'subgrism64'
    readout = 'nope'
    ngroup = 90
    obs_time = 6.0
    nint, exp_time = jwst.bin_search_exposure_time(
        inst, subarray, readout, ngroup, obs_time,
    )
    assert exp_time == 0.0


@pytest.mark.skip(reason='TBD')
def test_integration_time():
    pass


def test_extract_flux_rate_perform_calculation_single():
    with open('mocks/perform_calculation_nircam_lw_tsgrism.pkl', 'rb') as f:
        result = pickle.load(f)

    pixel_rate, full_well = jwst.extract_flux_rate(result)
    expected_rate = 1300.2144775390625
    expected_well = 58100.00422843957
    np.testing.assert_allclose(pixel_rate, expected_rate)
    np.testing.assert_allclose(full_well, expected_well)


def test_extract_flux_rate_perform_calculation_multi():
    with open('mocks/perform_calculation_miri_mrs_ts.pkl', 'rb') as f:
        result = pickle.load(f)

    pixel_rate, full_well = jwst.extract_flux_rate(result)
    expected_rate = [
        204.37779236, 109.17165375,  32.75593948,   4.55676889,
    ]
    expected_well = [
        193654.99564498, 193655.00540975, 193655.004117  , 193654.99798687,
    ]
    np.testing.assert_allclose(pixel_rate, expected_rate)
    np.testing.assert_allclose(full_well, expected_well)


def test_extract_flux_rate_tso_calculation_single():
    with open('mocks/tso_calculation_nircam_lw_tsgrism.pkl', 'rb') as f:
        tso = pickle.load(f)

    pixel_rate, full_well = jwst.extract_flux_rate(tso)
    expected_rate = [1261.23925781, 1300.21447754]
    expected_well = [58100.00016363, 58100.00129109]
    np.testing.assert_allclose(pixel_rate, expected_rate)
    np.testing.assert_allclose(full_well, expected_well)


def test_extract_flux_rate_tso_calculation_multi():
    with open('mocks/tso_calculation_miri_mrs_ts.pkl', 'rb') as f:
        tso = pickle.load(f)

    pixel_rate, full_well = jwst.extract_flux_rate(tso)
    expected_rate = [
        198.24740601, 105.90003204,  31.80119514,   4.46783495,
        204.37779236, 109.17165375,  32.75593948,   4.55676889,
    ]
    expected_well = [
        193654.99147199, 193654.99886718, 193654.98493823, 193654.99820043,
        193654.99564498, 193655.00540975, 193655.004117  , 193654.99798687,
    ]
    np.testing.assert_allclose(pixel_rate, expected_rate)
    np.testing.assert_allclose(full_well, expected_well)


@pytest.mark.skip(reason='TBD')
def test_extract_flux_rate_tso_calculation_get_max():
    pass


@pytest.mark.skip(reason='TBD')
def test_extract_flux_rate():
    pass


@pytest.mark.skip(reason='TBD')
def test_estimate_flux_rate():
    pass


@pytest.mark.skip(reason='TBD')
def test_groups_below_saturation():
    pass


@pytest.mark.skip(reason='TBD')
def test_get_sed_list():
    pass


@pytest.mark.skip(reason='TBD')
def test_find_closest_sed():
    pass


@pytest.mark.skip(reason='TBD')
def test_find_nearby_seds():
    pass


@pytest.mark.skip(reason='TBD')
def test_make_scene():
    pass


@pytest.mark.skip(reason='TBD')
def test_extract_sed():
    pass


@pytest.mark.skip(reason='TBD')
def test_blackbody_eclipse_depth():
    pass


@pytest.mark.skip(reason='TBD')
def test_set_depth_scene():
    pass


def test_get_bandwidths_nirspec_spectro():
    inst = 'nircam'
    mode = 'lw_ts'
    aper = 'lw'
    filter = 'f250m'
    wl0, bw, min_wl, max_wl = jwst.get_bandwidths(inst, mode, aper, filter)
    np.testing.assert_allclose(wl0, 2.5032492904902526)
    np.testing.assert_allclose(bw, 0.18093647144369174)
    np.testing.assert_allclose(min_wl, 2.4099999999997896)
    np.testing.assert_allclose(max_wl, 2.5949999999997693)


def test_get_bandwidths_miri_photo():
    inst = 'miri'
    mode = 'imaging_ts'
    aper = 'imager'
    filter = 'f560w'
    wl0, bw, min_wl, max_wl = jwst.get_bandwidths(inst, mode, aper, filter)
    np.testing.assert_allclose(wl0, 5.635006830413706)
    np.testing.assert_allclose(bw, 1.0006342537921697)
    np.testing.assert_allclose(min_wl, 5.029999999999501)
    np.testing.assert_allclose(max_wl, 6.204999999999371)


@pytest.mark.skip(reason='TBD')
def test_save_tso():
    pass


def test_simulate_tso_spectroscopy():
    pando = jwst.PandeiaCalculation('nircam', 'lw_tsgrism')
    pando.set_scene('phoenix', 'k5v', '2mass,ks', 8.351)

    disperser='grismr'
    filter='f444w'
    readout='rapid'
    subarray='subgrism64'
    ngroup = 14

    transit_dur = 2.71
    obs_dur = 6.01
    obs_type = 'transit'
    model_path = f'{ROOT}data/models/WASP80b_transit.dat'
    depth_model = np.loadtxt(model_path, unpack=True)

    tso = pando.tso_calculation(
        obs_type, transit_dur, obs_dur, depth_model,
        ngroup, disperser, filter, subarray, readout,
    )

    sim = jwst.simulate_tso(tso, n_obs=1, resolution=50.0, noiseless=True)
    bin_wl, bin_spec, bin_err, bin_widths = sim

    #print(' '.join([f'{val:.10e},' for val in bin_spec]))
    #print(' '.join([f'{val:.10e},' for val in bin_err]))
    expected_wl = [
        3.76373232, 3.83976732, 3.91733838, 3.99647653, 4.07721343,
        4.15958137, 4.24361332, 4.32934288, 4.41680435, 4.50603273,
        4.59706369, 4.68993366, 4.7846798 , 4.88134   , 4.9642892,
    ]
    expected_depths = [
        2.9196147533e-02, 2.9133292403e-02, 2.9080336314e-02, 2.9072502418e-02,
        2.9099839002e-02, 2.9181665646e-02, 2.9334965459e-02, 2.9401305002e-02,
        2.9380707088e-02, 2.9308734016e-02, 2.9250582642e-02, 2.9310178172e-02,
        2.9331133145e-02, 2.9383420209e-02, 2.9416676413e-02,
    ]
    expected_errors = [
        2.5216620191e-03, 9.7522860111e-05, 3.0456858027e-05, 2.8706507253e-05,
        2.9208368639e-05, 2.9868139512e-05, 3.0767740663e-05, 3.3243328518e-05,
        3.5617857052e-05, 3.8147765558e-05, 4.0227910565e-05, 4.3083460408e-05,
        4.5791244642e-05, 4.9872653719e-05, 7.4568353154e-05,    ]
    np.testing.assert_allclose(bin_wl, expected_wl)
    np.testing.assert_allclose(bin_spec, expected_depths)
    np.testing.assert_allclose(bin_err, expected_errors)

    # In case you want to see:
    if False:
        import matplotlib.pyplot as plt
        plt.figure(10)
        plt.clf()
        plt.plot(depth_model[0], depth_model[1], c='salmon')
        plt.xlim(2.0, 15)
        plt.errorbar(
            bin_wl, bin_spec, yerr=bin_err, xerr=bin_widths,
            fmt='o', color='xkcd:blue', mec='k', mew=1.0,
        )
        plt.xlim(3.7, 5.1)
        plt.ylim(0.0287, 0.0298)


def test_simulate_tso_photometry_nircam():
    pando = jwst.PandeiaCalculation('nircam', 'lw_ts')
    scene = jwst.make_scene('phoenix', 'k5v', '2mass,ks', 13.5)
    pando.calc['scene'] = [scene]

    filters = pando.get_configs('filters')
    subarray = 'sub160p'
    ngroup = 6
    model_path = f'{ROOT}data/models/WASP80b_transit.dat'
    depth_model = np.loadtxt(model_path, unpack=True)

    transit_dur = 2.71
    obs_dur = 6.0
    obs_type = 'transit'

    photo = []
    depths = np.zeros(len(filters))
    errors = np.zeros(len(filters))
    wl0 = np.zeros(len(filters))
    for i,filter in enumerate(filters):
        tso = pando.tso_calculation(
            obs_type, transit_dur, obs_dur, depth_model,
            ngroup, filter=filter, subarray=subarray,
        )
        sim = jwst.simulate_tso(tso, n_obs=1, noiseless=True)
        bin_wl, bin_spec, bin_err, bin_widths = sim
        wl0[i] = bin_wl[0]
        depths[i] = bin_spec[0]
        errors[i] = bin_err[0]
        photo.append(sim)

    #print(' '.join([f'{val:.10e},' for val in depths]))
    #print(' '.join([f'{val:.10e},' for val in errors]))
    expected_wl = [
       2.50043641, 2.74658655, 2.98917098, 3.10777101, 3.23666867,
       3.35436687, 3.52803572, 3.61482427, 4.05288935, 4.07232472,
       4.2764497 , 4.33293163, 4.62826398, 4.65428414, 4.70788841, 4.8098832
    ]
    expected_depths = [
        2.9602338269e-02, 2.9673371519e-02, 2.9704446669e-02, 2.9605682107e-02,
        2.9750487589e-02, 2.9726817356e-02, 2.9526279939e-02, 2.9459430393e-02,
        2.9076228245e-02, 2.9157602907e-02, 2.9342234464e-02, 2.9243180723e-02,
        2.9275985608e-02, 2.9279524748e-02, 2.9312241769e-02, 2.9349643103e-02,
    ]
    expected_errors = [
        1.3840358487e-04, 7.4326836871e-05, 1.2111463701e-04, 5.6637595030e-05,
        5.8597305401e-04, 1.2942307791e-04, 8.9283843176e-05, 1.4150529160e-04,
        7.6607387286e-04, 1.6205714300e-04, 2.6322708714e-04, 1.1749776751e-04,
        3.5311120708e-04, 1.3152440611e-03, 1.4757125324e-03, 3.2309278153e-04,
    ]
    np.testing.assert_allclose(wl0, expected_wl)
    np.testing.assert_allclose(depths, expected_depths)
    np.testing.assert_allclose(errors, expected_errors)

    # In case you want to see:
    if False:
        import matplotlib.pyplot as plt
        plt.figure(10)
        plt.clf()
        plt.plot(depth_model[0], depth_model[1], c='0.6')
        plt.xlim(2.3, 6.0)
        plt.ylim(0.0283, 0.0304)
        for i,sim in enumerate(photo):
            bin_wl, bin_spec, bin_err, bin_widths = sim
            col = plt.cm.rainbow(i/len(filters))
            plt.errorbar(
                bin_wl, bin_spec, yerr=bin_err, xerr=bin_widths.T/2,
                fmt='o', color=col, mec='k', mew=1.0, label=filters[i],
            )
        plt.legend(loc='best')



def test_simulate_tso_photometry_miri():
    # With MIRI / imaging
    pando = jwst.PandeiaCalculation('miri', 'imaging_ts')
    scene = jwst.make_scene('phoenix', 'k5v', '2mass,ks', 13.5)
    pando.calc['scene'] = [scene]

    model_path = f'{ROOT}data/models/WASP80b_transit.dat'
    depth_model = np.loadtxt(model_path, unpack=True)
    transit_dur = 2.71
    obs_dur = 6.0
    obs_type = 'transit'
    ngroup = 6
    photo = []
    filters = ['f560w', 'f770w', 'f1000w', 'f1130w', 'f1280w']

    depths = np.zeros(len(filters))
    errors = np.zeros(len(filters))
    wl0 = np.zeros(len(filters))
    for i,filter in enumerate(filters):
        tso = pando.tso_calculation(
            obs_type, transit_dur, obs_dur, depth_model,
            ngroup, filter=filter,
        )
        sim = jwst.simulate_tso(tso, n_obs=1, noiseless=True)
        bin_wl, bin_spec, bin_err, bin_widths = sim
        wl0[i] = bin_wl[0]
        depths[i] = bin_spec[0]
        errors[i] = bin_err[0]
        photo.append(sim)

    #print(' '.join([f'{val:.10e},' for val in depths]))
    #print(' '.join([f'{val:.10e},' for val in errors]))
    expected_wl = 5.60136086,  7.53420869,  9.88160076, 11.2961649 , 12.70594686
    expected_depths = [
        2.9531964334e-02, 2.9716731220e-02, 2.9197822880e-02,
        2.9140090022e-02, 2.9250074765e-02,
    ]
    expected_errors = [
        2.5261903335e-04, 2.5860955286e-04, 5.1512651158e-04,
        1.6119916090e-03, 1.1052531125e-03,
    ]
    np.testing.assert_allclose(wl0, expected_wl)
    np.testing.assert_allclose(depths, expected_depths)
    np.testing.assert_allclose(errors, expected_errors)

    # In case you want to see:
    if False:
        import matplotlib.pyplot as plt
        plt.figure(10)
        plt.clf()
        plt.plot(depth_model[0], depth_model[1], c='0.6')
        plt.xlim(2.0, 15)
        for i,sim in enumerate(photo):
            bin_wl, bin_spec, bin_err, bin_widths = sim
            col = plt.cm.rainbow(i/len(filters))
            plt.errorbar(
                bin_wl, bin_spec, yerr=bin_err, xerr=bin_widths.T/2,
                fmt='o', color=col, mec='k', mew=1.0, label=filters[i],
            )
        plt.legend(loc='best')


def test__print_pandeia_exposure_config():
    with open('mocks/perform_calculation_nircam_lw_tsgrism.pkl', 'rb') as f:
        result = pickle.load(f)
    config = result['input']['configuration']
    text = jwst._print_pandeia_exposure(config=config)
    assert text == 'Exposure time: 701.41 s (0.19 h)'


def test__print_pandeia_exposure_vals():
    inst = 'nircam'
    subarray = 'subgrism64'
    readout = 'rapid'
    ngroup = 90
    nint = 150
    text = jwst._print_pandeia_exposure(inst, subarray, readout, ngroup, nint)
    assert text == 'Exposure time: 4650.09 s (1.29 h)'


@pytest.mark.parametrize('format', [None, 'html', 'rich'])
def test__print_pandeia_exposure_format(format):
    with open('mocks/perform_calculation_nircam_lw_tsgrism.pkl', 'rb') as f:
        result = pickle.load(f)
    config = result['input']['configuration']
    text = jwst._print_pandeia_exposure(config=config, format=format)
    assert text == 'Exposure time: 701.41 s (0.19 h)'


def test__print_pandeia_saturation_perform_calc():
    # HERE!
    with open('mocks/perform_calculation_nircam_lw_tsgrism.pkl', 'rb') as f:
        result = pickle.load(f)

    text = jwst._print_pandeia_saturation(reports=[result], format=None)
    expected_text = """Max fraction of saturation: 1.5%
ngroup below 80% saturation: 104
ngroup below 100% saturation: 131"""
    assert text == expected_text


def test__print_pandeia_saturation_tso_calc():
    with open('mocks/tso_calculation_miri_mrs_ts.pkl', 'rb') as f:
        tso = pickle.load(f)

    text = jwst._print_pandeia_saturation(reports=tso, format=None)
    expected_text = """Max fraction of saturation: 73.2%
ngroup below 80% saturation: 273
ngroup below 100% saturation: 341"""
    assert text == expected_text


def test__print_pandeia_saturation_values():
    pixel_rate, full_well = 1396.5528564453125, 58100.002280515175
    inst = 'nircam'
    subarray = 'subgrism64'
    readout = 'rapid'
    ngroup = 90
    text = jwst._print_pandeia_saturation(
        inst, subarray, readout, ngroup, pixel_rate, full_well,
        format=None,
    )
    expected_text = """Max fraction of saturation: 73.7%
ngroup below 80% saturation: 97
ngroup below 100% saturation: 122"""
    assert text == expected_text


def test__print_pandeia_saturation_req_saturation():
    pixel_rate, full_well = 1396.5528564453125, 58100.002280515175
    inst = 'nircam'
    subarray = 'subgrism64'
    readout = 'rapid'
    ngroup = 90
    text = jwst._print_pandeia_saturation(
        inst, subarray, readout, ngroup, pixel_rate, full_well,
        format=None,
        req_saturation=70.0,
    )
    expected_text = """Max fraction of saturation: 73.7%
ngroup below 70% saturation: 85
ngroup below 100% saturation: 122"""
    assert text == expected_text


@pytest.mark.skip(reason='TBD')
def test__print_pandeia_stats():
    pass


def test__print_pandeia_report_perform_calculation_single():
    with open('mocks/perform_calculation_nircam_lw_tsgrism.pkl', 'rb') as f:
        result = pickle.load(f)

    report = jwst._print_pandeia_report([result], format=None)
    expected_report = """Exposure time: 701.41 s (0.19 h)
Max fraction of saturation: 1.5%
ngroup below 80% saturation: 104
ngroup below 100% saturation: 131

Signal-to-noise ratio        363.4
Extracted flux              1815.8  e-/s
Flux standard deviation        5.0  e-/s
Brightest pixel rate        1300.2  e-/s

Integrations:                         683
Duty cycle:                          0.66
Total exposure time:                701.4  s
First--last dt per exposure:        701.4  s
Reset--last dt per integration:     232.6  s

Reference wavelength:                    4.46  microns
Area of extraction aperture:             4.76  pixels
Area of background measurement:           6.3  pixels
Background surface brightness:            0.3  MJy/sr
Total sky flux in background aperture:   4.94  e-/s
Total flux in background aperture:      59.59  e-/s
Background flux fraction from scene:     0.92
Number of cosmic rays:      0.0002  events/pixel/read"""
    assert report == expected_report



def test__print_pandeia_report_perform_calculation_multi():
    with open('mocks/perform_calculation_miri_mrs_ts.pkl', 'rb') as f:
        result = pickle.load(f)

    report = jwst._print_pandeia_report(result, format=None)
    expected_report = """Exposure time: 27858.63 s (7.74 h)
Max fraction of saturation: 73.2%
ngroup below 80% saturation: 273
ngroup below 100% saturation: 341

Signal-to-noise ratio       1334.2    1265.2     960.1     179.2
Extracted flux               225.1     147.4      46.2       2.0  e-/s
Flux standard deviation        0.2       0.1       0.0       0.0  e-/s
Brightest pixel rate         204.4     109.2      32.8       4.6  e-/s

Integrations:                          40
Duty cycle:                          0.99
Total exposure time:              27858.6  s
First--last dt per exposure:      27858.6  s
Reset--last dt per integration:   27639.4  s

Reference wavelength:                    5.35   8.15  12.50  19.29  microns
Area of extraction aperture:             8.20   5.21   2.98   1.61  pixels
Area of background measurement:          79.4   50.5   28.9   15.6  pixels
Background surface brightness:            1.0   10.9   43.0  165.5  MJy/sr
Total sky flux in background aperture:   0.29   0.82   2.98   2.44  e-/s
Total flux in background aperture:       0.85   1.47   3.19   2.54  e-/s
Background flux fraction from scene:     0.66   0.45   0.06   0.04
Number of cosmic rays:      0.3122  events/pixel/read"""
    assert report == expected_report


def test__print_pandeia_report_tso_calculation_single():
    with open('mocks/tso_calculation_nircam_lw_tsgrism.pkl', 'rb') as f:
        result = pickle.load(f)

    report = jwst._print_pandeia_report(result, format=None)
    expected_report = """Exposure time: 21607.44 s (6.00 h)
Max fraction of saturation: 68.6%
ngroup below 80% saturation: 104
ngroup below 100% saturation: 131

Signal-to-noise ratio       3241.5
Extracted flux              1761.3  e-/s
Flux standard deviation        0.5  e-/s
Brightest pixel rate        1261.2  e-/s

                               in-transit  out-transit
Integrations:                         244      453
Duty cycle:                          0.99     0.99
Total exposure time:               7564.2  14043.3  s
First--last dt per exposure:       7564.2  14043.3  s
Reset--last dt per integration:    7396.7  13732.4  s

Reference wavelength:                    4.46  microns
Area of extraction aperture:             4.76  pixels
Area of background measurement:           6.3  pixels
Background surface brightness:            0.3  MJy/sr
Total sky flux in background aperture:   4.94  e-/s
Total flux in background aperture:      57.95  e-/s
Background flux fraction from scene:     0.91
Number of cosmic rays:      0.0072  events/pixel/read"""
    assert report == expected_report


def test__print_pandeia_report_tso_calculation_multi():
    with open('mocks/tso_calculation_miri_mrs_ts.pkl', 'rb') as f:
        tso = pickle.load(f)

    report = jwst._print_pandeia_report(tso, format=None)
    expected_report = """Exposure time: 20893.28 s (5.80 h)
Max fraction of saturation: 73.2%
ngroup below 80% saturation: 273
ngroup below 100% saturation: 341

Signal-to-noise ratio        657.0     623.0     472.3      87.5
Extracted flux               218.4     142.9      44.8       1.9  e-/s
Flux standard deviation        0.3       0.2       0.1       0.0  e-/s
Brightest pixel rate         198.2     105.9      31.8       4.5  e-/s

                               in-transit  out-transit
Integrations:                          10       20
Duty cycle:                          0.99     0.99
Total exposure time:               6962.6  13927.9  s
First--last dt per exposure:       6962.6  13927.9  s
Reset--last dt per integration:    6909.8  13819.7  s

Reference wavelength:                    5.35   8.15  12.50  19.29  microns
Area of extraction aperture:             8.20   5.21   2.98   1.61  pixels
Area of background measurement:          79.4   50.5   28.9   15.6  pixels
Background surface brightness:            1.0   10.9   43.0  165.5  MJy/sr
Total sky flux in background aperture:   0.29   0.82   2.98   2.44  e-/s
Total flux in background aperture:       0.83   1.45   3.18   2.54  e-/s
Background flux fraction from scene:     0.65   0.44   0.06   0.04
Number of cosmic rays:      0.3122  events/pixel/read"""
    assert report == expected_report


def test_tso_print_plain(capsys):
    with open('mocks/tso_calculation_nircam_lw_tsgrism.pkl', 'rb') as f:
        result = pickle.load(f)
    # Make some troublesome figures
    result['report_in']['scalar']['duty_cycle'] = 0.25
    result['report_in']['input']['configuration']['detector']['ngroup'] = 130
    result['report_out']['scalar']['duty_cycle'] = 0.25
    result['report_out']['input']['configuration']['detector']['ngroup'] = 130

    jwst.tso_print(result, format=None)
    captured = capsys.readouterr()
    expected_report = (
        "Exposure time: 31103.65 s (8.64 h)\r\n"
        "Max fraction of saturation: 99.1%\r\n"
        "ngroup below 80% saturation: 104\r\n"
        "ngroup below 100% saturation: 131\r\n"
        "\r\n"
        "Signal-to-noise ratio       3241.5\r\n"
        "Extracted flux              1761.3  e-/s\r\n"
        "Flux standard deviation        0.5  e-/s\r\n"
        "Brightest pixel rate        1261.2  e-/s\r\n"
        "\r\n"
        "                               in-transit  out-transit\r\n"
        "Integrations:                         244      453\r\n"
        "Duty cycle:                          0.25     0.25\r\n"
        "Total exposure time:               7564.2  14043.3  s\r\n"
        "First--last dt per exposure:       7564.2  14043.3  s\r\n"
        "Reset--last dt per integration:    7396.7  13732.4  s\r\n"
        "\r\n"
        "Reference wavelength:                    4.46  microns\r\n"
        "Area of extraction aperture:             4.76  pixels\r\n"
        "Area of background measurement:           6.3  pixels\r\n"
        "Background surface brightness:            0.3  MJy/sr\r\n"
        "Total sky flux in background aperture:   4.94  e-/s\r\n"
        "Total flux in background aperture:      57.95  e-/s\r\n"
        "Background flux fraction from scene:     0.91\r\n"
        "Number of cosmic rays:      0.0072  events/pixel/read\r\n"
    )
    assert captured.out == expected_report


@pytest.mark.skip(reason='Having some issues with capsys IO')
def test_tso_print_rich(capsys):
    with open('mocks/tso_calculation_nircam_lw_tsgrism.pkl', 'rb') as f:
        result = pickle.load(f)
    # Make some troublesome figures
    result['report_in']['scalar']['duty_cycle'] = 0.25
    result['report_in']['input']['configuration']['detector']['ngroup'] = 130
    result['report_out']['scalar']['duty_cycle'] = 0.25
    result['report_out']['input']['configuration']['detector']['ngroup'] = 130

    jwst.tso_print(result)
    captured = capsys.readouterr()
    expected_report = (
        "Exposure time: 31014.40 s (8.62 h)\r\n"
        "Max fraction of saturation: 106.4%\r\n"
        "ngroup below 80% saturation: 97\r\n"
        "ngroup below 100% saturation: 122\r\n"
        "\r\n"
        "Signal-to-noise ratio       3484.8\r\n"
        "Extracted flux              2043.0  e-/s\r\n"
        "Flux standard deviation        0.6  e-/s\r\n"
        "Brightest pixel rate        1354.7  e-/s\r\n"
        "\r\n"
        "                               in-transit  out-transit\r\n"
        "Integrations:                         243      452\r\n"
        "Duty cycle:                          0.25     0.25\r\n"
        "Total exposure time:               7533.2  14012.3  s\r\n"
        "First--last dt per exposure:       7533.2  14012.3  s\r\n"
        "Reset--last dt per integration:    7366.4  13702.1  s\r\n"
        "\r\n"
        "Reference wavelength:                    4.36  microns\r\n"
        "Area of extraction aperture:             4.76  pixels\r\n"
        "Area of background measurement:           6.3  pixels\r\n"
        "Background surface brightness:            0.3  MJy/sr\r\n"
        "Total sky flux in background aperture:   4.45  e-/s\r\n"
        "Total flux in background aperture:      64.12  e-/s\r\n"
        "Background flux fraction from scene:     0.93\r\n"
        "Number of cosmic rays:      0.0072  events/pixel/read\r\n"
    )
    assert captured.out == expected_report







# Copyright (c) 2025 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)
 
import gen_tso.utils as u
from prompt_toolkit.formatted_text import FormattedText


def test_format_text_plain():
    text = 'WASP-80 b'
    formatted = u.format_text(text, danger=True)
    assert formatted == text


def test_format_text_normal():
    text = 'WASP-80 b'
    formatted = u.format_text(text, warning=False, danger=False, format='html')
    assert formatted == text


def test_format_text_html_warning():
    text = 'WASP-80 b'
    formatted = u.format_text(text, warning=True, format='html')
    assert formatted == '<span class="warning">WASP-80 b</span>'


def test_format_text_html_danger():
    text = 'WASP-80 b'
    formatted = u.format_text(text, danger=True, format='html')
    assert formatted == '<span class="danger">WASP-80 b</span>'


def test_format_text_rich_warning():
    text = 'WASP-80 b'
    formatted = u.format_text(text, warning=True, format='rich')
    assert formatted == '<warning>WASP-80 b</warning>'


def test_format_text_rich_danger():
    text = 'WASP-80 b'
    formatted = u.format_text(text, danger=True, format='rich')
    assert formatted == '<danger>WASP-80 b</danger>'


def test_format_text_warning():
    text = 'WASP-80 b'
    formatted = u.format_text(text, warning=True, format='html')
    assert formatted == '<span class="warning">WASP-80 b</span>'


def test_format_text_danger1():
    text = 'WASP-80 b'
    formatted = u.format_text(text, danger=True, format='html')
    assert formatted == '<span class="danger">WASP-80 b</span>'


def test_format_text_danger2():
    text = 'WASP-80 b'
    formatted = u.format_text(text, warning=True, danger=True, format='html')
    assert formatted == '<span class="danger">WASP-80 b</span>'


def test_read_spectrum_file_success():
    pass


def test_read_spectrum_file_fail_none():
    pass


def test_read_spectrum_file_fail_warning_format():
    pass


def test_read_spectrum_file_fail_warning_wrong_file():
    pass


def test_read_spectrum_file_fail_error_format():
    pass


def test_read_spectrum_file_fail_error_wrong_file():
    pass


def test_collect_spectra_success():
    pass


def test_collect_spectra_success_and_fail_none():
    pass


def test_collect_spectra_success_and_fail_warning():
    pass


def test_collect_spectra_fail_error():
    pass



# Copyright (c) 2025 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)

import pytest
import re
import numpy as np
import gen_tso.pandeia_io as jwst
from gen_tso.utils import ROOT


def test_perform_calculation():
    pando = jwst.PandeiaCalculation('niriss', 'soss')
    # Set the stellar scene:
    pando.set_scene('phoenix', 'k5v', '2mass,ks', 8.637)

    # Set a NIRSpec observation
    disperser = 'gr700xd'
    filter = 'clear'
    readout = 'nisrapid'
    subarray = 'substrip96'
    aperture = 'soss'
    ngroup = 3
    nint = 984

    report = pando.perform_calculation(
        ngroup, nint, disperser, filter, subarray, readout, aperture,
    )
    # TBD: asserts


#def test_read_noise_variance():
#    report_config = {
#        'mode': '',
#        'aperture': '',
#    }
#    inputs = (
#        ('miri', 'mrs_ts', 'ch1'),
#        ('miri', 'lrsslitless', 'imager'),
#        ('nircam', 'lw_tsgrism', 'lw'),
#        ('nirspec', 'bots',  's1600a1'),
#        ('niriss', 'soss', 'soss'),
#    )
#    for input in inputs:
#        inst, mode, aperture = input
#        ins_config = get_instrument_config('jwst', inst)
#        pando = jwst.PandeiaCalculation(inst, mode)
#        #print(pando.calc['configuration']['instrument']['aperture'])
#        info_dict = ins_config['detector_config']
#        report_config['mode'] = mode
#        report_config['aperture'] = aperture
#
#        #report_config = report['input']['configuration']['instrument']
#        if report_config['mode'] == 'mrs_ts':
#            aperture = report_config['aperture']
#            aperture = ins_config['aperture_config'][aperture]['detector']
#
#        if aperture not in ins_config['detector_config']:
#            aperture = 'default'
#
#        read_noise = ins_config['detector_config'][aperture]['rn']
#        if isinstance(read_noise, dict):
#            read_noise = read_noise['default']
#        print(inst, mode, read_noise)

def test_get_scene_phoenix():
    pando = jwst.PandeiaCalculation('nirspec', 'bots')
    pando.set_scene(
        sed_type='phoenix', sed_model='k5v',
        norm_band='2mass,ks', norm_magnitude=8.351,
    )
    scene = pando.get_scene()
    expected_scene = {
        'sed_type': 'phoenix',
        'key': 'k5v',
        'normalization': 'photsys',
        'bandpass': '2mass,ks',
        'norm_flux': 8.351,
        'norm_fluxunit': 'vegamag',
    }
    for key, val in expected_scene.items():
        assert scene[key] == val


def test_get_scene_unset():
    pando = jwst.PandeiaCalculation('nirspec', 'bots')
    scene = pando.get_scene()
    expected_scene = {
        'sed_type': 'flat',
        'unit': 'fnu',
        'z': 0.0,
        'normalization': 'at_lambda',
        'norm_wave': 2.0,
        'norm_flux': 0.001,
        'norm_fluxunit': 'mjy',
        'norm_waveunit': 'microns',
    }
    for key, val in expected_scene.items():
        assert scene[key] == val


# Spectroscopy
def test_get_configs_miri_lrs(capsys):
    pando = jwst.PandeiaCalculation('miri', 'lrsslitless')
    pando.get_configs()

    captured = capsys.readouterr()
    expected_captured = """apertures: ['imager']
dispersers: ['p750l']
filters: ['']
subarrays: ['slitlessprism']
readout patterns: ['fastr1', 'slowr1']
"""


def test_get_configs_miri_mrs(capsys):
    pando = jwst.PandeiaCalculation('miri', 'mrs_ts')
    pando.get_configs()

    captured = capsys.readouterr()
    expected_captured = """apertures: ['ch1', 'ch2', 'ch3', 'ch4']
dispersers: ['short', 'medium', 'long']
filters: ['']
subarrays: ['full']
readout patterns: ['fastr1', 'slowr1']
"""


def test_get_configs_nircam_lw_tsgrism(capsys):
    pando = jwst.PandeiaCalculation('nircam', 'lw_tsgrism')
    pando.get_configs()

    captured = capsys.readouterr()
    expected_captured = """apertures: ['lw']
dispersers: ['grismr']
filters: ['f277w', 'f322w2', 'f356w', 'f444w']
subarrays: ['full', 'subgrism128', 'subgrism256', 'subgrism64', 'full (noutputs=1)', 'subgrism128 (noutputs=1)', 'subgrism256 (noutputs=1)', 'subgrism64 (noutputs=1)', 'sub41s1_2-spectra', 'sub82s2_4-spectra', 'sub164s4_8-spectra', 'sub260s4_8-spectra']
readout patterns: ['rapid', 'bright1', 'bright2', 'shallow2', 'shallow4', 'medium2', 'medium8', 'mediumdeep2', 'mediumdeep8', 'deep2', 'deep8', 'dhs3', 'dhs4', 'dhs5', 'dhs6', 'dhs7']
"""


def test_get_configs_nircam_sw_tsgrism(capsys):
    pando = jwst.PandeiaCalculation('nircam', 'sw_tsgrism')
    pando.get_configs()

    captured = capsys.readouterr()
    expected_captured = """apertures: ['dhs0spec2', 'dhs0spec4', 'dhs0spec8', 'dhs0bright']
dispersers: ['dhs0']
filters: ['f070w', 'f090w', 'f115w', 'f150w', 'f150w2', 'f200w']
subarrays: ['sub41s1_2-spectra', 'sub82s2_4-spectra', 'sub164s4_8-spectra', 'sub260s4_8-spectra']
readout patterns: ['rapid', 'bright1', 'dhs3', 'dhs4', 'dhs5', 'dhs6', 'dhs7']
"""


def test_get_configs_nirspec_bots(capsys):
    pando = jwst.PandeiaCalculation('nirspec', 'bots')
    pando.get_configs()

    captured = capsys.readouterr()
    expected_captured = """apertures: ['s1600a1']
grating/filter pairs: ['g140h/f070lp', 'g140h/f100lp', 'g140m/f070lp', 'g140m/f100lp', 'g235h/f170lp', 'g235m/f170lp', 'g395h/f290lp', 'g395m/f290lp', 'prism/clear']
subarrays: ['sub512', 'sub512s', 'sub1024a', 'sub1024b', 'sub2048']
readout patterns: ['nrs', 'nrsrapid']
"""


def test_get_configs_niriss_soss(capsys):
    pando = jwst.PandeiaCalculation('niriss', 'soss')
    pando.get_configs()

    captured = capsys.readouterr()
    expected_captured = """apertures: ['soss']
dispersers: ['gr700xd']
filters: ['clear', 'f277w']
subarrays: ['substrip256', 'substrip96', 'sossfull']
readout patterns: ['nis', 'nisrapid']
"""

# Acquisition
def test_get_configs_miri_acq():
    pando = jwst.PandeiaCalculation('miri', 'target_acq')
    pando.get_configs()

    captured = capsys.readouterr()
    expected_captured = """apertures: ['imager']
dispersers: []
filters: ['f560w', 'f1000w', 'f1500w', 'fnd']
subarrays: ['full', 'brightsky', 'sub256', 'sub128', 'sub64', 'slitlessprism']
readout patterns: ['fast', 'fastgrpavg', 'fastgrpavg8', 'fastgrpavg16', 'fastgrpavg32', 'fastgrpavg64']
"""


def test_get_configs_nircam_acq():
    pando = jwst.PandeiaCalculation('nircam', 'target_acq')
    pando.get_configs()

    captured = capsys.readouterr()
    expected_captured = """apertures: ['lw']
dispersers: []
filters: ['f335m', 'f405n']
subarrays: ['sub32tats']
readout patterns: ['rapid', 'bright1', 'bright2', 'shallow2', 'shallow4', 'medium2', 'medium8', 'mediumdeep2', 'mediumdeep8', 'deep2', 'deep8']
"""


def test_get_configs_nirspec_acq():
    pando = jwst.PandeiaCalculation('nirspec', 'target_acq')
    pando.get_configs()

    captured = capsys.readouterr()
    expected_captured = """apertures: ['s1600a1']
dispersers: []
filters: ['f110w', 'f140x', 'clear']
subarrays: ['full', 'sub32', 'sub2048']
readout patterns: ['nrsrapid', 'nrsrapidd6']
"""


def test_get_configs_niriss_acq():
    pando = jwst.PandeiaCalculation('niriss', 'target_acq')
    pando.get_configs()

    captured = capsys.readouterr()
    expected_captured = """apertures: ['imager', 'nrm']
dispersers: []
filters: ['f480m']
subarrays: ['subtaami']
readout patterns: ['nis', 'nisrapid']
"""


def test_show_config(capsys):
    pando = jwst.PandeiaCalculation('nirspec', 'bots')
    pando.set_scene(
        sed_type='phoenix', sed_model='k5v',
        norm_band='2mass,ks', norm_magnitude=8.351,
    )
    pando.show_config()

    captured = capsys.readouterr()
    expected_captured = """Instrument configuration:
    instrument = 'nirspec'
    mode = 'bots'
    aperture = 's1600a1'
    disperser = 'g395h'
    filter = 'f290lp'
    readout pattern = 'nrsrapid'
    subarray = 'sub2048'

Scene configuration:
    sed_type = 'phoenix'
    key = 'k5v'
    normalization = 'photsys'
    bandpass = '2mass,ks'
    norm_flux = 8.351
    norm_fluxunit = 'vegamag'
"""
    assert captured.out == expected_captured


def test_show_config_soss(capsys):
    pando = jwst.PandeiaCalculation('niriss', 'soss')
    pando.set_scene(
        sed_type='phoenix', sed_model='k5v',
        norm_band='2mass,ks', norm_magnitude=8.351,
    )
    pando.show_config()

    captured = capsys.readouterr()
    expected_captured = """Instrument configuration:
    instrument = 'niriss'
    mode = 'soss'
    aperture = 'soss'
    disperser = 'gr700xd'
    filter = 'clear'
    readout pattern = 'nisrapid'
    subarray = 'substrip256'
    order = 1

Scene configuration:
    sed_type = 'phoenix'
    key = 'k5v'
    normalization = 'photsys'
    bandpass = '2mass,ks'
    norm_flux = 8.351
    norm_fluxunit = 'vegamag'
"""
    assert captured.out == expected_captured


def test_saturation_fraction_get_ngroup():
    pando = jwst.PandeiaCalculation('nircam', 'lw_tsgrism')
    pando.set_scene(
        sed_type='phoenix', sed_model='k5v',
        norm_band='2mass,ks', norm_magnitude=8.351,
    )
    ngroup = pando.saturation_fraction(fraction=80.0)
    assert ngroup == 104


def test_saturation_fraction_get_saturation():
    pando = jwst.PandeiaCalculation('nircam', 'lw_tsgrism')
    pando.set_scene(
        sed_type='phoenix', sed_model='k5v',
        norm_band='2mass,ks', norm_magnitude=8.351,
    )
    saturation = pando.saturation_fraction(ngroup=91)

    expected_saturation = 69.3642895274301
    np.testing.assert_almost_equal(saturation, expected_saturation)


def test_saturation_fraction_with_flux_rate():
    pando = jwst.PandeiaCalculation('nircam', 'lw_tsgrism')
    pando.set_scene(
        sed_type='phoenix', sed_model='k5v',
        norm_band='2mass,ks', norm_magnitude=8.351,
    )
    flux_rate = 1400.2094016444962
    full_well = 58100.001867429375
    ngroup = pando.saturation_fraction(
        fraction=80.0, flux_rate=flux_rate, full_well=full_well,
    )
    assert ngroup == 97


def test_saturation_fraction_error_both_inputs():
    pando = jwst.PandeiaCalculation('nircam', 'lw_tsgrism')
    pando.set_scene(
        sed_type='phoenix', sed_model='k5v',
        norm_band='2mass,ks', norm_magnitude=8.351,
    )

    error = re.escape('Only one of fraction and ngroup must be defined')
    with pytest.raises(ValueError, match=error):
        saturation = pando.saturation_fraction(ngroup=91, fraction=80.0)


def test_saturation_fraction_error_no_inputs():
    pando = jwst.PandeiaCalculation('nircam', 'lw_tsgrism')
    pando.set_scene(
        sed_type='phoenix', sed_model='k5v',
        norm_band='2mass,ks', norm_magnitude=8.351,
    )

    error = re.escape('At least one of fraction and ngroup must be defined')
    with pytest.raises(ValueError, match=error):
        saturation = pando.saturation_fraction()


def test_saturation_fraction_no_guess_band(capsys):
    pando = jwst.PandeiaCalculation('nircam', 'lw_tsgrism')
    pando.set_scene(
        sed_type='phoenix', sed_model='k5v',
        norm_band='2mass,h', norm_magnitude=8.351,
    )

    ngroup = pando.saturation_fraction(fraction=80.0)
    assert ngroup is None
    captured = capsys.readouterr()
    expected_captured = 'Error, can only guess for Ks band ("2mass,ks")\n'
    assert captured.out == expected_captured


def test_saturation_fraction_no_guess_sed_type(capsys):
    pando = jwst.PandeiaCalculation('nircam', 'lw_tsgrism')
    ngroup = pando.saturation_fraction(fraction=80.0)
    assert ngroup is None
    captured = capsys.readouterr()
    expected_captured = 'Error, can only guess for phoenix or kurucz SEDs\n'
    assert captured.out == expected_captured


def test_saturation_fraction_no_guess_label(capsys):
    pando = jwst.PandeiaCalculation('nircam', 'lw_tsgrism')
    pando.set_scene(
        sed_type='phoenix', sed_model='k5v',
        norm_band='2mass,ks', norm_magnitude=8.351,
    )
    pando.calc['configuration']['instrument']['filter'] = 'bad_filter'

    ngroup = pando.saturation_fraction(fraction=80.0)
    assert ngroup is None
    captured = capsys.readouterr()
    expected_captured = "Error, no flux_rate spline for configuration label: 'lw_tsgrism_bad_filter_phoenix_k5v'\n"
    assert captured.out == expected_captured



def test_tso_calculation_single():
    pando = jwst.PandeiaCalculation('nirspec', 'bots')
    # Set the stellar scene:
    pando.set_scene('phoenix', 'k5v', '2mass,ks', 8.637)
    model_file = f'{ROOT}data/models/WASP80b_transit.dat'
    depth_model = np.loadtxt(model_file, unpack=True)

    # Set a NIRSpec observation
    disperser = 'g395h'
    filter = 'f290lp'
    readout = 'nrsrapid'
    subarray = 'sub2048'
    ngroup = 16

    transit_dur = 2.753
    obs_dur = 7.1
    obs_type = 'transit'
    tso = pando.tso_calculation(
        obs_type, transit_dur, obs_dur, depth_model,
        ngroup, disperser, filter, subarray, readout,
    )
    # assert


def test_tso_calculation_multiple():
    pando = jwst.PandeiaCalculation('miri', 'mrs_ts')
    pando.set_scene('phoenix', 'k5v', '2mass,ks', 8.637)
    model_file = f'{ROOT}data/models/WASP80b_transit.dat'
    depth_model = np.loadtxt(model_file, unpack=True)

    disperser = 'short'
    filter = None
    readout = 'fastr1'
    subarray = 'full'
    ngroup = 16
    aperture = ['ch1', 'ch2']

    transit_dur = 2.753
    obs_dur = 7.1
    obs_type = 'transit'
    tso = pando.tso_calculation(
        obs_type, transit_dur, obs_dur, depth_model,
        ngroup, disperser, filter, subarray, readout, aperture,
    )
    # assert


def test_calc_saturation_single():
    instrument = 'nircam'
    mode = 'lw_tsgrism'
    pando = jwst.PandeiaCalculation(instrument, mode)
    pando.set_scene('phoenix', 'k2v', '2mass,ks', 8.351)
    pixel_rate, full_well = pando.get_saturation_values(
        disperser='grismr', filter='f444w',
        readout='rapid', subarray='subgrism64',
    )
    np.testing.assert_almost_equal(pixel_rate, 1243.0863037109375)
    np.testing.assert_almost_equal(full_well, 58100.00)


def test_calc_saturation_multiple():
    inst = 'miri'
    mode = 'mrs_ts'
    pando = jwst.PandeiaCalculation(inst, mode)    
    pando.set_scene('phoenix', 'k2v', '2mass,ks', 8.351)
    ngroup = 2
    disperser = 'short'
    filter = None
    subarray = 'full'
    readout = 'fastr1'
    aperture = 'ch1 ch2 ch3 ch4'.split()

    pixel_rate, full_well = pando.get_saturation_values(
        disperser, filter, subarray, readout, ngroup, aperture,
    )
    expected_rate = [163.1600647,  89.8821335,  28.5365124,   4.2614131]
    expected_well = [193655.0, 193655.0, 193655.0, 193655.0]
    np.testing.assert_almost_equal(pixel_rate, expected_rate)
    np.testing.assert_almost_equal(full_well, expected_well)




# Copyright (c) 2025 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)

__all__ = [
    'find_target',
    'Catalog',
    'load_trexolists',
    'load_programs',
    'load_targets',
    'load_aliases',
    '_group_by_target',
]

import csv
from datetime import datetime
import json

from astropy.io import ascii
import numpy as np
import prompt_toolkit as ptk

from astropy.coordinates import Angle, SkyCoord
from astropy.units import hourangle, deg

from ..utils import ROOT
from . import utils as u
from .target import Target


def find_target(targets=None):
    """
    Interactive prompt with tab-completion to search for targets.

    Parameters
    ----------
    targets: list of Target objects

    Examples
    --------
    >>> import gen_tso.catalogs as cat
    >>> target = cat.find_target()
    """
    if targets is None:
        targets = load_targets('nea_data.txt', is_confirmed=True)
        targets += load_targets('tess_data.txt', is_confirmed=False)
    planets = [target.planet for target in targets]
    aliases = []
    for target in targets:
        aliases += target.aliases
    planets += list(aliases)

    completer = ptk.completion.WordCompleter(
        planets,
        sentence=True,
        match_middle=True,
    )
    session = ptk.PromptSession(
        history=ptk.history.FileHistory(f'{ROOT}/data/target_search_history')
    )
    name = session.prompt(
        "(Press 'tab' for autocomplete)\nEnter Planet name: ",
        completer=completer,
        complete_while_typing=False,
    )
    if name in aliases:
        for target in targets:
            if name in target.aliases:
                return target
    if name in planets:
        return targets[planets.index(name)]

    return None


class Catalog():
    """
    Load the entire catalog.

    Examples
    --------
    >>> import gen_tso.catalogs as cat
    >>> catalog = cat.Catalog()
    """
    def __init__(self, custom_targets=None):
        # Confirmed planets and TESS candidates
        nea_targets = load_targets('nea_data.txt', is_confirmed=True)
        tess_targets = load_targets('tess_data.txt', is_confirmed=False)
        self.targets = nea_targets + tess_targets
        if custom_targets is not None:
            custom = load_targets(custom_targets, is_confirmed=True)
            self.targets += custom

        # TBD: a switch between load_trexolists() and load_programs()?
        programs = load_trexolists(grouped=True)
        njwst = len(programs)
        host_aliases = load_aliases('host')

        jwst_hosts = []
        for jwst_target in programs:
            host_names = [obs['target'] for obs in jwst_target]
            nea_host = np.unique([
                host_aliases[host] if host in host_aliases else host
                for host in host_names
            ])
            jwst_hosts += list(nea_host)
            for obs in jwst_target:
                obs['nea_host'] = nea_host
        jwst_hosts = np.unique(jwst_hosts)

        planet_aliases = load_aliases('planet')
        planets_aka = u.invert_aliases(planet_aliases)

        for target in self.targets:
            target.is_jwst_host = target.host in jwst_hosts
            if target.is_jwst_host:
                for j in range(njwst):
                    if target.host == programs[j][0]['nea_host']:
                        break
                target.programs = programs[j]
                planets = []
                for obs in programs[j]:
                    planets += obs['planets']
                planets = np.unique(planets)
                letter = u.get_letter(target.planet).strip()
                target.is_jwst_planet = letter in planets
            else:
                target.is_jwst_planet = False

            if target.planet in planets_aka:
                target.aliases = planets_aka[target.planet]

        self._transit_mask = [target.is_transiting for target in self.targets]
        self._jwst_mask = [target.is_jwst_host for target in self.targets]
        self._confirmed_mask = [target.is_confirmed for target in self.targets]


    def get_target(
            self, name=None,
            is_transit=True, is_jwst=None, is_confirmed=None,
        ):
        """
        Search by name for a planet in the catalog.

        Parameters
        ----------
        name: String
            If not None, name of the planet to search.
            If None, an interactive prompt will open to search for the planet
        is_transit: Bool
            If True/False restrict search to transiting/non-transiting planets
            If None, consider all targets.
        is_jwst: Bool
            If True/False restrict search to planet of/not JWST hosts
            If None, consider all targets.
        is_confirmed: Bool
            If True/False restrict search to confirmed/candidate planets
            If None, consider all targets.

        Returns
        -------
        target: a Target object
            Target with the system properties of the searched planet.
            If no target was found on the catalog, return None.
        """
        mask = np.ones(len(self.targets), bool)
        if is_transit is not None:
            mask &= np.array(self._transit_mask) == is_transit
        if is_jwst is not None:
            mask &= np.array(self._jwst_mask) == is_jwst
        if is_confirmed is not None:
            mask &= np.array(self._confirmed_mask) == is_confirmed

        targets = [target for target,flag in zip(self.targets,mask) if flag]

        if name is None:
            return find_target(targets)

        target = u.normalize_name(name)
        for target in targets:
            if name == target.planet or name in target.aliases:
                return target

    def show_target(
            self, name=None,
            is_transit=True, is_jwst=None, is_confirmed=True,
        ):
        target = self.get_target(name, is_transit, is_jwst, is_confirmed)
        if target is None:
            return
        print(target)


def load_targets(database='nea_data.txt', is_confirmed=np.nan):
    """
    Unpack star and planet properties from plain text file.

    Parameters
    ----------
    database: String
        nea_data.txt or tess_data.txt

    Returns
    -------
    targets: List of Target

    Examples
    --------
    >>> import gen_tso.catalogs as cat
    >>> nea_data = cat.load_nea_targets_table()
    """
    with open(f'{ROOT}data/{database}', 'r') as f:
        lines = f.readlines()

    lines = [
        line for line in lines
        if not line.strip().startswith('#')
    ]
    targets = []
    for line in lines:
        if line.startswith('>'):
            name_len = line.find(':')
            host = line[1:name_len]
            star_vals = np.array(line[name_len+1:].split(), float)
            ra, dec, ks_mag, rstar, mstar, teff, logg, metal = star_vals
        elif line.startswith(' '):
            name_len = line.find(':')
            planet = line[1:name_len].strip()
            planet_vals = np.array(line[name_len+1:].split(), float)
            t_dur, rplanet, mplanet, sma, period, teq, min_mass = planet_vals

            target = Target(
                host=host,
                mstar=mstar, rstar=rstar, teff=teff, logg_star=logg,
                metal_star=metal,
                ks_mag=ks_mag, ra=ra, dec=dec,
                planet=planet,
                mplanet=mplanet, rplanet=rplanet,
                period=period, sma=sma, transit_dur=t_dur,
                is_confirmed=is_confirmed,
                is_min_mass=bool(min_mass),
            )
            targets.append(target)

    return targets


def _add_planet_info(observations):
    """
    Add planet letter info to a list of observations
    and other corrections
    """
    planets_file = f'{ROOT}data/programs/planets_per_program.json'
    with open(planets_file, "r") as f:
        planet_data = json.load(f)

    known_obs = []
    for obs in observations:
        pid = obs['pid']
        obs_id = obs['observation']
        key = f'{pid}_{obs_id}'
        known_obs.append(key)
        if key in planet_data:
            for var, value in planet_data[key].items():
                obs[var] = value

    for key,obs in planet_data.items():
        if key not in known_obs and 'missing' in obs:
            obs.pop('missing')
            date_format = "%Y-%m-%d %H:%M:%S"
            val = obs['date_start']
            if isinstance(val, str):
                obs['date_start'] = datetime.strptime(val, date_format)
            val = obs['date_end']
            if isinstance(val, str):
                obs['date_end'] = datetime.strptime(val, date_format)
            observations.append(obs)

    return observations


def _group_by_target(observations):
    """
    Group observations by host, using RA and dec to detect aliases
    for a same object
    """
    ra = [Angle(obs['ra'], unit=hourangle).deg for obs in observations]
    dec = [Angle(obs['dec'], unit=deg).deg for obs in observations]
    coords = SkyCoord(ra, dec, unit='deg', frame='icrs')

    nobs = len(observations)
    taken = np.zeros(nobs, bool)
    group_indices = []
    for i in range(nobs):
        if taken[i]:
            continue
        seps = coords[i].separation(coords).to('arcsec').value
        indices = np.where(seps < 50)[0]
        taken[indices] = True
        group_indices.append(indices)

    grouped_data = []
    for i,indices in enumerate(group_indices):
        target = [observations[j] for j in indices]
        grouped_data.append(target)

    return grouped_data


def load_trexolists(grouped=False, trexo_file=None):
    """
    Extract the JWST programs' data from a trexolists.csv file.
    Note that trexolists know targets by their host star, not by
    individual planets in a given system.

    Parameters
    ----------
    grouped: Bool
        - If False, return a single 1D list with observations.
        - If True, return a nested list of the observations per target.
    trexo_file: String
        If None, extract data from default Gen TSO location.
        Otherwise, a path to a trexolists.csv file.

    Returns
    -------
    observations: 1D or 2D list of dictionaries
        A list of all JWST observations, where each item is a dictionary
        containing the observation's details.
        If grouped is True, the output is a nested list, where the
        observations are grouped per target (host).

    Examples
    --------
    >>> import gen_tso.catalogs as cat
    >>>
    >>> # Get data as lists of individual programs:
    >>> observations = cat.load_trexolists()
    >>> # Show one of them:
    >>> observations[98]
    {'category': 'ERS',
     'pi': 'Batalha',
     'pid': '1366',
     'proprietary_period': 0,
     'target': 'WASP-39',
     'target_in_program': 'WASP-39',
     'observation': '3',
     'visit': '1',
     'status': 'Archived',
     'ra': '14:29:18.3955',
     'dec': '-03:26:40.20',
     'event': 'transit',
     'instrument': 'NIRSPEC',
     'mode': 'BOTS',
     'disperser': 'G395H',
     'filter': 'F290LP',
     'subarray': 'SUB2048',
     'readout': 'NRSRAPID',
     'groups': 70,
     'phase_start': 0.95248,
     'phase_end': 0.96275,
     'duration': 10.56,
     'plan_window': None,
     'date_start': datetime.datetime(2022, 7, 30, 20, 46, 32),
     'date_end': datetime.datetime(2022, 7, 31, 6, 21, 30),
     'planets': ['b']}

    >>> # Get data grouped per target (host):
    >>> observations = cat.load_trexolists(grouped=True)
    >>> for obs in observations[27]:
    >>>     print(f"{obs['pid']}  {obs['pi']}  {obs['instrument']} {obs['event']} ")
    1366  Batalha  NIRISS transit
    1366  Batalha  NIRCAM transit
    1366  Batalha  NIRSPEC transit
    1366  Batalha  NIRSPEC transit
    2783  Powell  MIRI transit
    5634  Baeyens  NIRSPEC see_phase
    """
    if trexo_file is None:
        trexo_file = f'{ROOT}data/trexolists.csv'

    trexolist_data = ascii.read(
        trexo_file,
        format='csv', guess=False, fast_reader=False, comment='#',
    )

    nirspec_filter = {
        'G395H': 'F290LP',
        'G395M': 'F290LP',
        'G235H': 'F170LP',
        'G235M': 'F170LP',
        'G140H': 'F100LP',
        'G140M': 'F100LP',
        'PRISM': 'CLEAR',
    }
    instrument = {
        'BOTS': 'NIRSPEC',
        'SOSS': 'NIRISS',
        'GTS': 'NIRCAM',
        'LRS': 'MIRI',
        'MRS': 'MIRI',
        'F1500W': 'MIRI',
        'F1280W': 'MIRI',
    }

    observations = []
    for i,data in enumerate(trexolist_data):
        obs = {}
        obs['category'] = str(data['ProposalCategory'])
        obs['pi'] = str(data['LastName'])
        obs['pid'] = str(data['ProposalID'])
        obs['cycle'] = str(data['Cycle'])
        obs['proprietary_period'] = int(data['ProprietaryPeriod'])

        target = str(data['hostname_nn'])
        obs['target'] = u.normalize_name(target)
        obs['target_in_program'] = target
        obs['planets'] = data['letter_nn'].split('+')
        obs['event'] = data['Event'].lower().replace('phasec', 'phase curve')

        obs['observation'] = str(data['Observation'])
        obs['visit'] = '1'
        obs['status'] = str(data['Status'])
        coordinates = data['EquatorialCoordinates'].split()
        obs['ra'] = ':'.join(coordinates[0:3])
        obs['dec'] = ':'.join(coordinates[3:6])

        mode = str(data['ObservingMode'])
        disperser = obs['disperser'] = str(data['GratingGrism'])
        inst = obs['instrument'] = instrument[mode]
        if mode == 'SOSS':
            disperser = 'None'
            filter = 'CLEAR'
        elif mode == 'LRS':
            disperser = 'None'
            filter = 'None'
        elif mode == 'MRS':
            disperser = 'unknown'
            filter = 'None'
        elif inst == 'MIRI':
            # disperser will be fixed below by _add_planet_info()
            disperser = 'None'
            filter = mode
            mode = 'Imaging TS'
        elif mode == 'GTS':
            mode = 'GRISMR TS'
            disperser, filter = disperser.split('+')
            if '_' in data['Subarray']:
                disperser = f'DHS0,{disperser}'
                # hard-coded, known up to Cycle4:
                # will be fixed below by _add_planet_info()
                filter = f'F150W2,{filter}'
        elif inst == 'NIRSPEC':
            filter = nirspec_filter[disperser]
        obs['mode'] = mode
        obs['disperser'] = disperser
        obs['filter'] = filter

        obs['subarray'] = str(data['Subarray'])
        obs['readout'] = str(data['ReadoutPattern'])
        obs['groups'] = int(data['Groups'])

        window = str(data['PlanWindow'])
        if window == 'X':
            obs['plan_window'] = None
        elif '(' in window:
            window = window[0:window.index('(')]
            w_start, w_end = window.split('-')
            start = datetime.strptime(w_start.strip(), '%b %d, %Y')
            end = datetime.strptime(w_end.strip(), '%b %d, %Y')
            obs['plan_window'] = f"{start.strftime('%Y-%m-%d')} - {end.strftime('%Y-%m-%d')}"
        else:
            obs['plan_window'] = window

        obs['duration'] = float(data['Hours'])

        date = data['StartTime']
        if date == 'X':
            obs['date_start'] = None
        else:
            obs['date_start'] = datetime.strptime(date, '%b %d, %Y %H:%M:%S')
        date = data['EndTime']
        if date == 'X':
            obs['date_end'] = None
        else:
            obs['date_end'] = datetime.strptime(date, '%b %d, %Y %H:%M:%S')

        observations.append(obs)

    observations = _add_planet_info(observations)

    if grouped:
        return _group_by_target(observations)
    return observations


def load_programs(grouped=False, csv_file=None):
    """
    Get the data from the downloaded JWST programs (xml files)
    Note that the programs know targets by host star, not by
    individual planets in a given system.

    Parameters
    ----------
    grouped: Bool
        - If False, return a 1D list of all observations
        - If True, return a nested list of observations grouped by
          host target.
    csv_file: String
        Path to a csv file saved with parse_programs().
        If None, load the default csv file of gen_tso (which should
        contain all known programs).

    Examples
    --------
    >>> import gen_tso.catalogs as cat
    >>> programs = cat.load_programs()
    """
    # Read the CSV file into a list of dictionaries
    if csv_file is None:
        csv_file = f'{ROOT}data/programs/jwst_tso_programs.csv'

    observations = []
    with open(csv_file, mode="r", encoding="utf-8") as file:
        reader = csv.DictReader(file)
        for row in reader:
            observations.append(row)

    # Parse data types
    int_keys = ['cycle', 'groups', 'integrations', 'proprietary_period']
    float_keys = ['duration', 'period', 'phase_start', 'phase_duration']
    date_keys = ['date_start', 'date_end']
    for obs in observations:
        for key,val in obs.items():
            if val == '':
                obs[key] = None
            elif key in int_keys:
                obs[key] = int(val)
            elif key in float_keys:
                obs[key] = float(val)
            elif key in date_keys:
                date_format = "%Y-%m-%d %H:%M:%S"
                obs[key] = datetime.strptime(val, date_format)
            elif key == 'planets':
                obs[key] = eval(obs[key])

    if grouped:
        return _group_by_target(observations)
    return observations


def parse(name, style):
    """
    Parse a planet name as a planet (no change) or as a star,
    in which case it will identify from confirmed or candidate
    format name.

    Parameters
    ----------
    style: String
        Select from 'planet' or 'host'.

    Returns
    -------
    name: String
        Parsed name.

    Examples
    --------
    >>> from gen_tso.catalogs.catalogs import parse
    >>> parse('WASP-80 b', 'planet')
    WASP-80 b'
    >>> parse('WASP-80 b', 'host')
    WASP-80'
    >>> parse('TOI-316.01', 'host')
    'TOI-316'
    """
    if style == 'planet':
        return name
    elif style == 'host':
        if u.is_letter(name):
            return name[:-2]
        end = name.rindex('.')
        return name[:end]


def load_aliases(style='planet', aliases_file=None):
    """
    Load file with known aliases of NEA targets.

    Parameters
    ----------
    style: String
        Select from 'planet', 'host', or 'system'.

    Returns
    -------
    aliases: Dictionary
        Dictionary of aliases to-from NASA Exoplanet Archive name.
        See below for examples depending on the 'style' argument.

    Examples
    --------
    >>> import gen_tso.catalogs as cat
    >>>
    >>> # From alias planet name to NEA name:
    >>> aliases = cat.load_aliases('planet')
    >>> aliases['CD-38 2551 b']
    'WASP-63 b'
    >>>
    >>> # From alias host name to NEA name:
    >>> aliases = cat.load_aliases('host')
    >>> aliases['CD-38 2551']
    'WASP-63'
    >>>
    >>> # As stellar system with all host and planet aliases:
    >>> aliases = cat.load_aliases('system')
    >>> aliases['WASP-63']
    {'host': 'WASP-63',
     'planets': ['WASP-63 b'],
     'host_aliases': array(['CD-38 2551', 'TOI-483', 'WASP-63'], dtype='<U10'),
     'planet_aliases': {'TOI-483.01': 'WASP-63 b',
      'CD-38 2551 b': 'WASP-63 b',
      'WASP-63 b': 'WASP-63 b'}}
    """
    if style not in ['planet', 'host', 'system']:
        raise ValueError(
            "Invalid alias style, select from: 'planet', 'host', or 'system'"
        )
    if aliases_file is None:
        aliases_file = f'{ROOT}data/target_aliases.txt'

    with open(aliases_file, 'r') as f:
        lines = f.readlines()

    if style != 'system':
        aliases = {}
        for line in lines:
            loc = line.index(':')
            name = parse(line[:loc], style)
            for alias in line[loc+1:].strip().split(','):
                aliases[parse(alias,style)] = name
            aliases[name] = name
        return aliases

    aliases = {}
    current_host = ''
    for line in lines:
        loc = line.index(':')
        planet = parse(line[:loc], 'planet')
        host = parse(line[:loc], 'host')
        host_aliases = [
            parse(name, 'host')
            for name in line[loc+1:].strip().split(',')
        ]
        host_aliases += [host]
        planet_aliases = {
            parse(name, 'planet'): planet
            for name in line[loc+1:].strip().split(',')
        }
        planet_aliases[planet] = planet
        if host != current_host:
            # Save old one
            if current_host != '':
                system['host_aliases'] = np.unique(system['host_aliases'])
                aliases[current_host] = system
            # Start new one
            system = {
                'host': host,
                'planets': [planet],
                'host_aliases': host_aliases,
                'planet_aliases': planet_aliases,
            }
            current_host = host
        else:
            system['host_aliases'] += host_aliases
            system['planets'] += [planet]
            system['planet_aliases'].update(planet_aliases)
    system['host_aliases'] = np.unique(system['host_aliases'])
    aliases[current_host] = system
    return aliases




# Copyright (c) 2025 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)

__all__ = [
    'read_noise_variance',
    'exposure_time',
    'bin_search_exposure_time',
    'integration_time',
    'saturation_level',
    'extract_flux_rate',
    'estimate_flux_rate',
    'groups_below_saturation',
    'get_sed_list',
    'find_closest_sed',
    'find_nearby_seds',
    'make_scene',
    'extract_sed',
    'blackbody_eclipse_depth',
    'set_depth_scene',
    'get_bandwidths',
    'save_tso',
    'simulate_tso',
    '_get_tso_wl_range',
    '_get_tso_depth_range',
    '_print_pandeia_exposure',
    '_print_pandeia_saturation',
    '_print_pandeia_stats',
    '_print_pandeia_report',
    'tso_print',
]

from collections.abc import Iterable
import copy
from decimal import Decimal
import json
import pickle
import random

import numpy as np
import scipy.interpolate as si
import pandeia.engine.sed as sed
from pandeia.engine.calc_utils import get_instrument_config
from pandeia.engine.normalization import NormalizationFactory
from pyratbay.spectrum import constant_resolution_spectrum, bin_spectrum
from pyratbay.tools import u
import prompt_toolkit

from ..utils import format_text
from .pandeia_defaults import (
    _photo_modes,
    _load_flux_rate_splines,
    get_sed_types,
    get_throughputs,
    make_saturation_label,
)
sed_types = get_sed_types()


def read_noise_variance(report, ins_config):
    """
    Extract the read noise from instrument configuration data.

    Parameters
    ----------
    report: dict
        A pandeia perform_calculation() output.
    ins_config: dict
        A pandeia get_instrument_config() output for a given instrument.

    Returns
    -------
    read_noise: Float
        The instrumental read noise (electrons per pixel?)
    """
    report_config = report['input']['configuration']['instrument']
    if report_config['mode'] in ['sw_tsgrism', 'sw_ts']:
        aperture = report_config['aperture']
        noise = ins_config['detector_config']['sw']['rn']
        if aperture not in noise:
            aperture = 'default'
        read_noise = noise[aperture]
        return read_noise

    if report_config['mode'] in ['mrs_ts', 'lrsslit']:
        aperture = report_config['aperture']
        aperture = ins_config['aperture_config'][aperture]['detector']
    else:
        aperture = report_config['aperture']

    if aperture not in ins_config['detector_config']:
        aperture = 'default'

    read_noise = ins_config['detector_config'][aperture]['rn']
    if isinstance(read_noise, dict):
        read_noise = read_noise['default']

    return read_noise


def _exposure_time_function(instrument, subarray, readout, ngroup, nexp=1):
    """
    Return a callable that evaluates the exposure time as a function
    of nint.

    Parameters
    ----------
    instrument: String
        Which instruments (miri, nircam, niriss, or nirspec).
    subarray: String
        Subarray mode for the given instrument.
    readout: String
        Readout pattern mode for the given instrument.
    ngroup: Integeer
        Number of groups per integration.  Must be >= 2.
    nexp: Integer
        Number of exposures.

    Returns
    -------
    exp_time: Float
        Exposure time in seconds.

    """
    if isinstance(instrument, str):
        telescope = 'jwst'
        config = get_instrument_config(telescope, instrument)
    else:
        config = instrument.ins_config

    # When switching instrument, subarray and readout updates are not atomic
    if subarray not in config['subarray_config']['default']:
        return None
    if readout not in config['readout_pattern_config']:
        return None

    subarray_config = config['subarray_config']
    readout_config = config['readout_pattern_config'][readout]

    nframe = readout_config['nframe']
    ndrop2 = readout_config['ndrop2']
    tfffr = subarray_config['default'][subarray]['tfffr']
    tframe = subarray_config['default'][subarray]['tframe']
    has_tframe = (
        readout in subarray_config and
        subarray in subarray_config[readout]
    )
    if has_tframe:
        tframe = subarray_config[readout][subarray]["tframe"]
        tfffr = subarray_config[readout][subarray]["tfffr"]

    ndrop1 = ndrop3 = 0
    if "ndrop1" in readout_config:
        ndrop1 = readout_config["ndrop1"]
    if "ndrop3" in readout_config:
        ndrop3 = readout_config["ndrop3"]

    nreset1 = nreset2 = 1
    if "nreset1" in subarray_config["default"][subarray]:
        nreset1 = subarray_config["default"][subarray]["nreset1"]
        nreset2 = subarray_config["default"][subarray]["nreset2"]
    has_nreset = (
        readout in subarray_config and
        subarray in subarray_config[readout] and
        "nreset1" in subarray_config[readout][subarray]
    )
    if has_nreset:
        nreset1 = subarray_config[readout][subarray]["nreset1"]
        nreset2 = subarray_config[readout][subarray]["nreset2"]

    if "nreset1" in readout_config:
        nreset1 = readout_config["nreset1"]
        nreset2 = readout_config["nreset2"]

    def exp_time(nint):
        time = nexp * (
            tfffr * nint +
            tframe * (
                nreset1 + (nint-1) * nreset2 +
                nint * (ndrop1 + (ngroup-1) * (nframe + ndrop2) + nframe + ndrop3)
            )
        )
        return time
    return exp_time


def exposure_time(instrument, subarray, readout, ngroup, nint, nexp=1):
    """
    Calculate the exposure time for the given instrumental setup.
    Based on pandeia.engine.exposure.

    Parameters
    ----------
    instrument: String
        Which instruments (miri, nircam, niriss, or nirspec).
    subarray: String
        Subarray mode for the given instrument.
    readout: String
        Readout pattern mode for the given instrument.
    ngroup: Integeer
        Number of groups per integration.  Must be >= 2.
    nint: Integer
        Number of integrations.
    nexp: Integer
        Number of exposures.

    Returns
    -------
    exp_time: Float
        Exposure time in seconds.

    Examples
    --------
    >>> import gen_tso.pandeia_io as jwst

    >>> inst = 'nircam'
    >>> subarray = 'subgrism64'
    >>> readout = 'rapid'
    >>> nint = 1
    >>> ngroup = 90
    >>> exp_time = jwst.exposure_time(inst, subarray, readout, ngroup, nint)
    >>> print(exp_time)
    """
    exp_time = _exposure_time_function(
        instrument, subarray, readout, ngroup, nexp,
    )
    if exp_time is None:
        return 0.0
    return exp_time(nint)


def bin_search_exposure_time(
        instrument, subarray, readout, ngroup, obs_time, nexp=1,
    ):
    """
    Binary search for nint such that exp_time(nint) > obs_time

    Parameters
    ----------
    instrument: String
        Which instruments (miri, nircam, niriss, or nirspec).
    subarray: String
        Subarray mode for the given instrument.
    readout: String
        Readout pattern mode for the given instrument.
    ngroup: Integeer
        Number of groups per integration.  Must be >= 2.
    obs_time: Integer
        Total observation time to aim for (in hours).
    nexp: Integer
        Number of exposures.

    Returns
    -------
    nint: Integer
        Number of integrations to reach obs_time.
    exp_time: Float
        Exposure time in seconds.

    Examples
    --------
    >>> import gen_tso.pandeia_io as jwst

    >>> instrument = 'miri'
    >>> subarray = 'slitlessprism'
    >>> readout = 'fastr1'
    >>> ngroup = 30
    >>> obs_time = 6.0
    >>> nint, exp_time = jwst.bin_search_exposure_time(
    >>>     instrument, subarray, readout, ngroup, obs_time,
    >>> )
    >>> print(nint, exp_time)
    """
    exp_time = _exposure_time_function(
        instrument, subarray, readout, ngroup, nexp,
    )
    if exp_time is None:
        return 0, 0.0

    obs_time_sec = 3600 * obs_time
    n1 = 1
    t1 = exp_time(n1)
    if obs_time_sec < t1:
        return n1, t1

    n2 = int(obs_time_sec / t1)
    t2 = exp_time(n2)
    while t2 < obs_time_sec:
        n1 = n2
        n2 = n1*2
        t2 = exp_time(n2)
    while n2-n1 > 1:
        n = int(0.5*(n1+n2))
        t = exp_time(n)
        if obs_time_sec > t:
            n1 = n
        else:
            n2 = n

    return n2, exp_time(n2)


def integration_time(instrument, subarray, readout, ngroup):
    """
    Compute JWST's integration time for a given instrument configuration.
    Based on pandeia.engine.exposure.get_times()

    Examples
    --------
    >>> import gen_tso.pandeia_io as jwst

    >>> # Calculate integration time for a given instrument setup:
    >>> inst = 'nircam'
    >>> ngroup = 90
    >>> readout = 'rapid'
    >>> subarray = 'subgrism64'
    >>> integ_time = jwst.integration_time(inst, subarray, readout, ngroup)
    >>> print(integ_time)
    30.6549
    """
    if isinstance(instrument, str):
        telescope = 'jwst'
        config = get_instrument_config(telescope, instrument)
    else:
        config = instrument.ins_config

    # When switching instrument, subarray and readout updates are not atomic
    if subarray not in config['subarray_config']['default']:
        return 0.0
    if readout not in config['readout_pattern_config']:
        return 0.0

    nframe = config['readout_pattern_config'][readout]['nframe']
    ndrop2 = config['readout_pattern_config'][readout]['ndrop2']
    tframe = config['subarray_config']['default'][subarray]['tframe']
    has_tframe = (
        readout in config['subarray_config'] and
        subarray in config['subarray_config'][readout]
    )
    if has_tframe:
        tframe = config['subarray_config'][readout][subarray]["tframe"]

    ndrop1 = 0
    if "ndrop1" in config['readout_pattern_config'][readout]:
        ndrop1 = config['readout_pattern_config'][readout]["ndrop1"]

    integration_time = tframe * (
        ndrop1 + (ngroup - 1) * (nframe + ndrop2) + nframe
    )
    return integration_time


def extract_flux_rate(reports, get_max=False):
    """
    Compute saturation values for a given perform_calculation output.

    Parameters
    ----------
    reports: Dictionary or list of dictionaries
        One or more pandeia's perform_calculation() output dictionary
        or a tso_calculation output dictionary.
        If there is more than one input report, return arrays of
        saturation values for each input.
    get_max: Bool
        If True and there is  more than one input report, return the
        saturation values for the report that's quickest to saturate.

    Returns
    -------
    brightest_pixel_rate: Float
        e- per second rate at the brightest pixel.
    full_well: Float
        Number of e- counts to saturate the detector.

    Examples
    --------
    >>> import gen_tso.pandeia_io as jwst
    >>>
    >>> inst = 'nircam'
    >>> readout = 'rapid'
    >>> pando = jwst.PandeiaCalculation(inst, 'lw_tsgrism')
    >>> pando.set_scene('phoenix', 'k5v', '2mass,ks', 8.351)
    >>> result = pando.perform_calculation(
    >>>     ngroup=2, nint=1, readout=readout, filter='f444w',
    >>> )
    >>> pixel_rate, full_well = jwst.extract_flux_rate(result)
    >>>
    >>> # Now I can calculate the saturation level for any integration time:
    >>> # (for the given filter and scene)
    >>> subarray = 'subgrism64'
    >>> for ngroup in [2, 97, 122]:
    >>>     integ_time = jwst.integration_time(inst, subarray, readout, ngroup)
    >>>     sat_level = 100 * pixel_rate * integ_time / full_well
    >>>     print(f'Sat. fraction for {ngroup:3d} groups: {sat_level:5.1f}%')
    Sat. fraction for   2 groups:   1.5%
    Sat. fraction for  97 groups:  73.9%
    Sat. fraction for 122 groups:  93.0%

    >>> # Calculate maximum number of groups before saturation
    >>> inst = 'nircam'
    >>> subarray = 'subgrism64'
    >>> readout = 'bright2'
    >>> pando = jwst.PandeiaCalculation(inst, 'lw_tsgrism')
    >>> pando.set_scene('phoenix', 'k5v', '2mass,ks', 8.351)
    >>> result = pando.perform_calculation(
    >>>     ngroup=2, nint=1, readout=readout, filter='f444w',
    >>> )
    >>> pixel_rate, full_well = jwst.extract_flux_rate(result)
    >>>
    >>> # ngroup staying below 80% of saturation:
    >>> req_fraction = 80.0
    >>> dt_integ = (
    >>>     jwst.integration_time(inst, subarray, readout, ngroup=3) -
    >>>     jwst.integration_time(inst, subarray, readout, ngroup=2)
    >>> )
    >>> sat_fraction = 100 * pixel_rate * dt_integ / full_well
    >>> ngroup_req = int(req_fraction/sat_fraction)
    >>> print(f'ngroup below {req_fraction:.1f}% of saturation: {ngroup_req}')
    ngroup below 80.0% of saturation: 52
    """
    if not isinstance(reports, list):
        reports = [reports]
    # Unpack TSO dictionary
    if 'report_in' in reports[0]:
        reports = (
            [report['report_in'] for report in reports] +
            [report['report_out'] for report in reports]
        )

    ncalc = len(reports)
    brightest_pixel_rate = np.zeros(ncalc)
    full_well = np.zeros(ncalc)
    for i,report in enumerate(reports):
        brightest_pixel_rate[i] = report['scalar']['brightest_pixel']
        full_well[i] = (
            brightest_pixel_rate[i]
            * report['scalar']['saturation_time']
            / report['scalar']['fraction_saturation']
        )

    if len(reports) == 1:
        return brightest_pixel_rate[0], full_well[0]
    if get_max:
        idx = np.argmax(brightest_pixel_rate*full_well)
        return brightest_pixel_rate[idx], full_well[idx]
    return brightest_pixel_rate, full_well


def saturation_level(reports, get_max=False):
    """
    Deprecated. Use `extract_flux_rate()` instead.
    """
    import warnings
    warnings.warn(
        "saturation_level() is deprecated and will be removed in a "
        "future release. Use extract_flux_rate() instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return extract_flux_rate(reports, get_max)


def estimate_flux_rate(
        sed_type, sed_model, ks_mag,
        mode, aperture, disperser, filter, subarray, order='',
    ):
    """
    Estimate a detector's brightest-pixel flux rate and the full well
    for a given source and an instrumental configuration,
    based on pre-tabulated Ks-bands magnitudes.

    Parameters
    ----------
    sed_type: String
        Type of model: 'phoenix', 'k93models', or 'bt_settl'.
    sed_model:
        The SED model required for each sed_type, see
        jwst.get_sed_list(), jwst.find_closest_sed() or jwst.find_nearby_seds()
    ks_mag: float
        Magnitude of the star in the Ks band.
    mode: String
        The observing mode for the JWST instrument.
    aperture: String
        Aperture configuration for the given instrument.
    disperser: String
        Disperser/grating for the given instrument.
    filter: String
        Filter for the given instrument.
    subarray: String
        Subarray mode for the given instrument.
    order: Integer
        For NIRISS SOSS only, the spectral order.

    Returns
    -------
    brightest_pixel_rate: Float
        e- per second rate at the brightest pixel.
    full_well: Float
        Number of e- counts to saturate the detector.

    Example
    -------
    >>> import gen_tso.pandeia_io as jwst

    >>> sed_type = 'phoenix'
    >>> sed_model = 'k5v'
    >>> ks_mag = 8.0
    >>> flux_rate, full_well = jwst.estimate_flux_rate(
    >>>     sed_type, sed_model, ks_mag,
    >>>     mode='lw_tsgrism', aperture='lw',
    >>>     disperser='grism', filter='f444w', subarray='subgrism64',
    >>> )
    >>> print(flux_rate, full_well)
    1796.190365030732 58100.0
    """
    sed_label = f'{sed_type}_{sed_model}'
    obs_label = make_saturation_label(
        mode, aperture, disperser, filter, subarray, order, sed_label,
    )
    flux_rate_spline, full_well = _load_flux_rate_splines(obs_label)

    if flux_rate_spline is None:
        return None, None

    pixel_rate = 10**flux_rate_spline(ks_mag)
    return pixel_rate, full_well


def groups_below_saturation(
        req_saturation,
        instrument=None, subarray=None, readout=None,
        flux_rate=None, full_well=None,
        mode=None, aperture=None, disperser=None, filter=None, order=None,
        sed_type=None, sed_model=None, ks_mag=None,
        reports=None,
    ):
    """
    Calculate the maximum number of groups below a give saturation level
    for the given configuration.

    Parameters
    ----------
    req_saturation: Float or 1D float iterable
        Required saturation level in percent.
    instrument: String
        Which instruments (miri, nircam, niriss, or nirspec).
    subarray: String
        Subarray mode for the given instrument.
    readout: String
        Readout pattern mode for the given instrument.
    flux_rate: Float
        e- per second rate at the brightest pixel.
    full_well: Float
        Number of e- counts to saturate the detector.
    mode: String
        The observing mode for the JWST instrument.
    aperture: String
        Aperture configuration for the given instrument.
    disperser: String
        Disperser/grating for the given instrument.
    filter: String
        Filter for the given instrument.
    order: Integer
        For NIRISS SOSS only, the spectral order.
    sed_type: String
        Type of SED model: 'phoenix', 'k93models', or 'bt_settl'.
    sed_model:
        The SED model required for each sed_type, see
        jwst.get_sed_list(), jwst.find_closest_sed() or jwst.find_nearby_seds()
    ks_mag: float
        Magnitude of the star in the Ks band.
    reports: Dictionary or list of dictionaries
        Either a pandeia's perform_calculation() output dictionary
        or a tso_calculation output dictionary.

    Returns
    -------
    ngroup: Integeer
        Number of groups per integration to keep the saturation fraction
        of the brightest pixel rate below req_saturation.

    Examples
    --------
    >>> import gen_tso.pandeia_io as jwst
    >>>
    >>> # Use a pandeia report as input:
    >>> pando = jwst.PandeiaCalculation('nircam', 'lw_tsgrism')
    >>> pando.set_scene('phoenix', 'k5v', '2mass,ks', 8.351)
    >>> report = pando.perform_calculation(ngroup=2, nint=1)
    >>> ngroup = jwst.groups_below_saturation([80, 100], reports=report)
    >>> print(ngroup)
    [104, 131]

    >>> # Use instrument and source variables as input:
    >>> inst = 'nircam'
    >>> mode = 'lw_tsgrism'
    >>> disperser = 'grismr'
    >>> filter = 'f444w'
    >>> subarray = 'subgrism64'
    >>> readout = 'rapid'
    >>> aperture = 'lw'
    >>>
    >>> sed_type = 'phoenix'
    >>> sed_model = 'k5v'
    >>> ks_mag = 8.351
    >>>
    >>> ngroup = jwst.groups_below_saturation(
    >>>     req_saturation=80.0,
    >>>     instrument=inst, mode=mode, disperser=disperser, filter=filter,
    >>>     subarray=subarray, readout=readout, aperture=aperture,
    >>>     sed_type=sed_type, sed_model=sed_model, ks_mag=ks_mag,
    >>>
    >>> )
    >>> print(ngroup)
    104
    """
    if isinstance(req_saturation, Iterable):
        req_saturations = req_saturation
    else:
        req_saturations = [req_saturation]

    if reports is not None:
        flux_rate, full_well = extract_flux_rate(reports, get_max=True)
        # Unpack instrumental configuration values:
        if not isinstance(reports, list):
            reports = [reports]
        if 'report_in' in reports[0]:
            report = reports[0]['report_in']
        else:
            report = reports[0]
        config = report['input']['configuration']
        instrument = config['instrument']['instrument']
        subarray = config['detector']['subarray']
        readout = config['detector']['readout_pattern']

    if flux_rate is None:
        flux_rate, full_well = estimate_flux_rate(
            sed_type, sed_model, ks_mag,
            mode, aperture, disperser, filter, subarray, order,
        )

    are_undefined = (
        instrument is None,
        subarray is None,
        readout is None,
        flux_rate is None,
        full_well is None,
    )
    if np.any(are_undefined):
        raise ValueError('Not all required inputs are defined')

    dt_1group = integration_time(instrument, subarray, readout, ngroup=1)
    dt_2group = integration_time(instrument, subarray, readout, ngroup=2)
    dt = dt_2group - dt_1group

    sat_1group = flux_rate * dt_1group / full_well
    sat_dt = flux_rate * dt / full_well

    ngroups = []
    for saturation in req_saturations:
        m = 1 + (saturation/100.0 - sat_1group) / sat_dt
        if np.isfinite(m):
            ngroups.append(int(m))
        else:
            ngroups.append(0)

    if not isinstance(req_saturation, Iterable):
        ngroups = ngroups[0]
    return ngroups


def get_sed_list(source):
    """
    Load list of available PHOENIX or Kurucz stellar SED models

    Parameters
    ----------
    source: String
        SED source: 'phoenix', 'k93models', or 'bt_settl'.

    Returns
    -------
    keys: list of strings
        SED model keys (as used in a Pandeia scene dict).
    names: list of strings
        SED model names (as seen on the ETC).
    teff: list of floats
        SED model effective-temperature grid.
    log_g: list of floats
        SED model log(g) grid.

    Examples
    --------
    >>> import gen_tso.pandeia_io as jwst
    >>> # Kurucz models
    >>> keys, names, teff, log_g = jwst.get_sed_list('k93models')
    >>> # PHOENIX models
    >>> keys, names, teff, log_g = jwst.get_sed_list('phoenix')
    >>> # BT-Settl models
    >>> keys, names, teff, log_g = jwst.get_sed_list('bt_settl')
    """
    sed_path = sed.default_refdata_directory
    with open(f'{sed_path}/sed/{source}/spectra.json', 'r') as f:
        info = json.load(f)
    names = np.array([model['display_string'] for model in info.values()])

    if source == 'bt_settl':
        teff = np.array([model.split()[1] for model in names], dtype=float)
        log_g = np.array([model.split()[2] for model in names], dtype=float)
    else:
        teff = np.array([model['teff'] for model in info.values()])
        log_g = np.array([model['log_g'] for model in info.values()])

    keys = np.array(list(info.keys()))
    tsort = np.argsort(teff)[::-1]

    return keys[tsort], names[tsort], teff[tsort], log_g[tsort]


def find_closest_sed(teff, logg, sed_type='phoenix'):
    """
    A very simple cost-function to find the closest stellar model
    within a non-regular Teff-log_g grid.

    Since these are not regular grids, the cost function is not an
    absolute science, it depends on what weights more Teff of logg
    for a given case.  The current formula seems to be a good balance.

    Parameters
    ----------
    teff: float
        Target effective temperature.
    logg: float
        Target log(g).
    sed_type: String
        Select from 'phoenix' or 'k93models'

    Returns
    -------
    sed: String
        The SED key that best matches the teff,logg pair.

    Examples
    --------
    >>> import gen_tso.pandeia_io as jwst
    >>>
    >>> # Kurucz models
    >>> sed = jwst.find_closest_sed(
    >>>     teff=4143.0, logg=4.66, sed_type='k93models',
    >>> )
    >>> print(f'SED: {repr(sed)}')
    SED: 'k7v'
    """
    keys, names, models_teff, models_logg = get_sed_list(sed_type)
    cost = (
        np.abs(np.log10(teff/models_teff)) +
        np.abs(logg-models_logg) / 15.0
    )
    idx = np.argmin(cost)
    return keys[idx]


def find_nearby_seds(teff, dt=200.0, sed_type='all'):
    """
    Show all SED models with temperatures in range teff +/- dt.
    Highlight the closest(s) one(s) to teff with asterisks.
    Return the parameters of the closest SED.

    Parameters
    ----------
    teff: float
        Target effective temperature.
    dt: float
        Temperature range around teff to explore.
    sed_type: String or iterable of strings.
        Select one from 'phoenix', 'k93models', or 'bt_settl'.
        Or, select 'all' to consider all SED types.
        Or, provide a list with the SED types to consider.

    Returns
    -------
    model: Tuple
        A tuple with the (sed type, sed key, temperature, log_g) of
        the model closes to teff  Note there can be multiple models at
        same distance from teff, only one is returned.

    Examples
    --------
    >>> import gen_tso.pandeia_io as jwst
    >>>
    >>> # Kurucz models
    >>> sed = jwst.find_nearby_seds(teff=5100.0, dt=700, sed_type='k93models')
    SED key      teff  logg  SED type
    -----------  ----------  -----------
    'k4v'        4560   4.5  'k93models'
    'g5i'        4850   1.1  'k93models'
    'k0v'        5250   4.5  'k93models'  **
    'g8v'        5570   4.5  'k93models'
    'g5v'        5770   4.5  'k93models'

    >>> # All models
    >>> sed = jwst.find_nearby_seds(teff=3600.0, dt=200, sed_type='all')
    SED key      teff  logg  SED type
    -----------  ----------  -----------
    'm3.5-3400'  3400   5.0  'bt_settl'
    'm2v'        3500   4.6  'k93models'  **
    'm2i'        3500   0.0  'phoenix'    **
    'm5v'        3500   5.0  'phoenix'    **
    'm2.5'       3500   5.0  'bt_settl'   **
    'm2v'        3500   4.5  'phoenix'    **
    'k5i'        3750   0.5  'phoenix'
    'm0v'        3750   4.5  'phoenix'
    'm0iii'      3750   1.5  'phoenix'
    'm0i'        3750   0.0  'phoenix'

    >>> # PHOENIX + BT-Settl models
    >>> sed = jwst.find_nearby_seds(
    >>>     teff=3320.0, dt=250, sed_type=['phoenix', 'bt_settl'],
    >>> )
    SED key      teff  logg  SED type
    -----------  ----------  -----------
    'm4.5-3100'  3100   5.0  'bt_settl'
    'm4.5-3200'  3200   5.0  'bt_settl'
    'm3.5-3300'  3300   5.0  'bt_settl'   **
    'm3.5-3400'  3400   5.0  'bt_settl'
    'm2.5'       3500   5.0  'bt_settl'
    'm5v'        3500   5.0  'phoenix'
    'm2i'        3500   0.0  'phoenix'
    'm2v'        3500   4.5  'phoenix'
    """
    sed_types = get_sed_types()
    if sed_type == 'any' or sed_type=='all':
        sed_type = sed_types
    elif isinstance(sed_type, str):
        sed_type = [sed_type]

    for type in sed_type:
        if type not in sed_types:
            raise ValueError(
                f'Invalid sed_type {repr(sed_type)}, select from {sed_types}'
            )

    models = []
    for sed_t in sed_type:
        keys, names, models_teff, logg = get_sed_list(sed_t)
        nmodels = len(keys)
        for i in range(nmodels):
            if np.abs(models_teff[i]-teff) <= dt:
                model = (str(sed_t), keys[i], models_teff[i], logg[i])
                models.append(model)

    temps = [model[2] for model in models]
    i_min = np.argmin(np.abs(np.array(temps)-teff))
    dt_min = np.abs(temps[i_min]-teff)
    closest_model = models[i_min]

    isort = np.argsort(temps)
    models = [models[i] for i in isort]
    print('SED key      teff  logg  SED type')
    print('-----------  ----------  -----------')
    for model in models:
         sed, key, temp, logg = model
         suff = '  **' if np.abs(temp-teff)==dt_min else ''
         print(f'{repr(str(key)):11}  {temp:4.0f}   {logg:.1f}  {repr(sed):11}{suff}')

    return closest_model


def make_scene(sed_type, sed_model, norm_band=None, norm_magnitude=None):
    """
    Create a stellar point-source scene dictionary for use in Pandeia.

    Parameters
    ----------
    sed_type: String
        Type of model:
        - 'phoenix', 'k93models',
        - 'blackbody', 'input', or 'flat'
    sed_model:
        The SED model required for each sed_type:
        - phoenix, k93models: the model key (see get_sed_list)
        - blackbody: the effective temperature (K)
        - input: dict with 'wl' and 'flux' keys containing the
          wavelength (un) and SED spectrum (mJy).
        - flat: the unit ('flam' or 'fnu')
    norm_band: String
        Band over which to normalize the spectrum.
    norm_magnitude: float
        Magnitude of the star at norm_band.

    Examples
    --------
    >>> import gen_tso.pandeia_io as jwst

    >>> sed_type = 'phoenix'
    >>> sed_model = 'k5v'
    >>> norm_band = '2mass,ks'
    >>> norm_magnitude = 8.637
    >>> scene = jwst.make_scene(sed_type, sed_model, norm_band, norm_magnitude)
    >>> print(scene)
    {'spectrum': {'sed': {'sed_type': 'phoenix', 'key': 'k5v'},
      'normalization': {'type': 'photsys',
       'bandpass': '2mass,ks',
       'norm_flux': 8.637,
       'norm_fluxunit': 'vegamag'},
      'extinction': {'bandpass': 'j',
       'law': 'mw_rv_31',
       'unit': 'mag',
       'value': 0},
      'lines': [],
      'redshift': 0},
     'position': {'orientation': 0.0, 'x_offset': 0.0, 'y_offset': 0.0},
     'shape': {'geometry': 'point'}}

    >>> sed_type = 'blackbody'
    >>> sed_model = 4250.0
    >>> norm_band = '2mass,ks'
    >>> norm_magnitude = 8.637
    >>> scene = jwst.make_scene(sed_type, sed_model, norm_band, norm_magnitude)
    """
    if sed_type == 'kurucz':
        sed_type = 'k93models'

    sed = {'sed_type': sed_type}

    if sed_type == 'flat':
        sed['unit'] = sed_model
    elif sed_type in sed_types:
        sed['key'] = sed_model
    elif sed_type == 'blackbody':
        sed['temp'] = sed_model
    elif sed_type == 'input':
        sed['spectrum'] = sed_model['wl'], sed_model['flux']

    if norm_band is None or norm_band == 'none':
        normalization = {'type': 'none'}
    else:
        normalization = {
            'type': 'photsys',
            'bandpass': norm_band,
            'norm_flux': norm_magnitude,
            'norm_fluxunit': 'vegamag',
        }

    spectrum = {
        'sed': sed,
        'normalization': normalization,
        'extinction': {
            'bandpass': 'j',
            'law': 'mw_rv_31',
            'unit': 'mag',
            'value': 0,
        },
        'lines': [],
        'redshift': 0,
    }

    scene = {
        'spectrum': spectrum,
        'position': {'orientation': 0.0, 'x_offset': 0.0, 'y_offset': 0.0},
        'shape': {'geometry': 'point'},
    }
    return scene


def extract_sed(scene, wl_range=None):
    """
    Extract the flux spectrum array from a given scene dict.

    Parameters
    ----------
    scene: dict
        A pandeia scene dict

    Returns
    -------
    wl: 1D float array
        Scene's wavelength array (um).
    flux: 1D float array
        Scene's flux spectrum (mJy).

    Examples
    --------
    >>> import gen_tso.pandeia_io as jwst
    >>> import matplotlib.pyplot as plt

    >>> sed_type = 'phoenix'
    >>> sed_model = 'k5v'
    >>> norm_band = '2mass,ks'
    >>> norm_magnitude = 8.637
    >>> scene1 = jwst.make_scene(sed_type, sed_model, norm_band, norm_magnitude)
    >>> wl1, phoenix = jwst.extract_sed(scene1)

    >>> sed_type = 'k93models'
    >>> sed_model = 'k7v'
    >>> scene2 = jwst.make_scene(sed_type, sed_model, norm_band, norm_magnitude)
    >>> wl2, kurucz = jwst.extract_sed(scene2)

    >>> sed_type = 'blackbody'
    >>> sed_model = 4250.0
    >>> scene3 = jwst.make_scene(sed_type, sed_model, norm_band, norm_magnitude)
    >>> wl3, bb = jwst.extract_sed(scene3)

    >>> plt.figure(0)
    >>> plt.clf()
    >>> plt.plot(wl1, phoenix, c='b')
    >>> plt.plot(wl2, kurucz, c='darkorange')
    >>> plt.plot(wl3, bb, c='xkcd:green')
    >>> plt.xlim(0.5, 12)
    """
    sed_model = sed.SEDFactory(config=scene['spectrum']['sed'], webapp=True, z=0)
    normalization = NormalizationFactory(
        config=scene['spectrum']['normalization'],
        webapp=True,
    )
    wave, flux = normalization.normalize(sed_model.wave, sed_model.flux)
    if normalization.type == 'none':
        if sed_model.sed_type in sed_types:
            # Convert wavelengths from A to um:
            wave *= 1e-4  # pc.A / pc.um
        if sed_model.sed_type == 'blackbody':
            # Convert blackbody intensity to flux:
            flux *= np.pi

    if wl_range is None:
        wl_mask = np.ones(len(wave.value), bool)
    else:
        wl_mask = (wave.value >= wl_range[0]) & (wave.value <= wl_range[1])
    return wave.value[wl_mask], flux.value[wl_mask]


def blackbody_eclipse_depth(t_planet, rprs, sed_type, sed_model, return_fluxes=False):
    """
    Compute an eclipse-depth spectrum assuming a blackbody planet emission.

    Parameters
    ----------
    t_planet: Float
        Planet effective temperature (kelvin).
    rprs: Float
        Planet-to-star radius ratio.
    sed_type: String
        Type of stellar SED model: select one from:
        'phoenix', 'k93models', 'blackbody', or 'input'.
    sed_model:
        The SED model required for each sed_type:
        - for phoenix or k93models: the model key (see get_sed_list)
        - for blackbody: the effective temperature (K)
        - for input: a dict including 'wl' and 'flux' keys containing the
              wavelength (um) and SED flux spectrum (mJy).
    return_fluxes: Bool
        If True, also return the planetary and stellar fluxes

    Returns
    -------
    wl: 1D float array
        Eclipse depth's wavelength array (um).
    depth: 1D float array
        Eclipse depth spectrum.
    f_planet: 1D float array
        Planet surface flux (erg s-1 cm-2 Hz-1).
    f_star: 1D float array
        Stellar surface flux (erg s-1 cm-2 Hz-1).

    Examples
    --------
    >>> import gen_tso.pandeia_io as jwst
    >>>
    >>> t_planet = 2000.0
    >>> rprs = 0.1
    >>> sed_type = 'phoenix'
    >>> sed_model = 'k5v'
    >>> wl, depth = jwst.blackbody_eclipse_depth(t_planet, rprs, sed_type, sed_model)
    """
    # Un-normalized planet and star SEDs
    star_scene = make_scene(sed_type, sed_model, norm_band='none')
    planet_scene = make_scene('blackbody', t_planet, norm_band='none')
    wl_star, f_star = extract_sed(star_scene)
    wl_planet, f_planet = extract_sed(planet_scene)
    # Interpolate black body at wl_star
    interp_func = si.interp1d(
        wl_planet, f_planet, bounds_error=False, fill_value=np.nan,
    )
    f_planet = interp_func(wl_star)
    wl_mask = np.isfinite(f_planet)
    wl = wl_star[wl_mask]
    # Eclipse_depth = Fplanet/Fstar * rprs**2
    depth = f_planet[wl_mask] / f_star[wl_mask] * rprs**2
    if return_fluxes:
        return wl, depth, f_planet[wl_mask], f_star[wl_mask]
    return wl, depth


def set_depth_scene(scene, obs_type, depth_model, wl_range=None):
    """
    Given a stellar point-source scene and an observing geometry,
    generate a new scene with the source flux scaled by a transit depth
        flux_in_transit = flux * (1-depth),
    or scaled by an eclipse depth:
        flux_out_of_eclipse = flux * (1+depth).

    To combine stellar and depth spectra, the lower-resolution
    spectrum will be linearly-interpolated to the sampling of the
    higher-resolution spectrum.

    Also return a scene of the un-scaled stellar source (with the same
    sampling as in the scaled scene, which prevents fringing when
    operating on the spectra produced from these scenes).
    Both output scenes are already normalized (so that a pandeia
    calculation does not undo the eclipse/transit depth scaling).

    Parameters
    ----------
    scene: A pandeia scene dictionary
        An input scene of a stellar point source.
    obs_type: string
        The observing geometry 'transit' or 'eclipse'.
    depth_model: list of two 1D array or a 2D array
        The transit or eclipse depth spectrum where the first item is
        the wavelength (um) and the second is the depth.
    wl_range: float iterable
        A wl_min, wl_max pair where to clip the stellar and depth spectra.

    Examples
    --------
    >>> import gen_tso.pandeia_io as jwst

    >>> # A pandeia stellar point-source scene:
    >>> scene = jwst.make_scene(
    >>>     sed_type='k93models', sed_model='k7v',
    >>>      norm_band='2mass,ks', norm_magnitude=8.637,
    >>> )
    >>>
    >>> # A transit-depth spectrum:
    >>> depth_model = np.loadtxt('WASP80b_transit.dat', unpack=True)
    >>>
    >>> obs_type = 'transit'
    >>> star_scene, in_transit_scene = jwst.set_depth_scene(
    >>>     scene, obs_type, depth_model,
    >>> )
    """
    # from pandeia_interface import extract_sed
    # Get stellar flux (mJy) from scene, planet depth spectrum from input
    wl_star, star_flux = extract_sed(scene)
    wl_planet, depth = depth_model

    if wl_range is not None:
        wl_mask = (wl_star > wl_range[0]) & (wl_star < wl_range[1])
        wl_star = wl_star[wl_mask]
        star_flux = star_flux[wl_mask]

        wl_mask = (wl_planet > wl_range[0]) & (wl_planet < wl_range[1])
        wl_planet = wl_planet[wl_mask]
        depth = depth[wl_mask]

    # Interpolate to the highest-resolution spectrum
    R_star = np.median(wl_star[1:]/np.abs(np.ediff1d(wl_star)))
    R_planet = np.median(wl_planet[1:]/np.abs(np.ediff1d(wl_planet)))

    if R_star > R_planet:
        wl = wl_star
        interp_func = si.interp1d(
            wl_planet, depth, bounds_error=False, fill_value=0.0,
        )
        depth = interp_func(wl)
    else:
        wl = wl_planet
        interp_func = si.interp1d(
            wl_star, star_flux, bounds_error=False, fill_value=0.0,
        )
        star_flux = interp_func(wl)

    if obs_type == 'transit':
        scaled_flux = star_flux * (1-depth)
    elif obs_type == 'eclipse':
        scaled_flux = star_flux * (1+depth)

    isort = np.argsort(wl)
    depth_scene = copy.deepcopy(scene)
    depth_scene['spectrum']['sed'] = dict(
        sed_type = 'input',
        spectrum = [wl[isort], scaled_flux[isort]],
    )
    # Do not normalize anymore, otherwise we lose the depth correction
    depth_scene['spectrum']['normalization'] = dict(type="none")

    # Now the star:
    star_scene = copy.deepcopy(scene)
    star_scene['spectrum']['sed'] = dict(
        sed_type = 'input',
        spectrum = [wl[isort], star_flux[isort]],
    )
    star_scene['spectrum']['normalization'] = dict(type="none")

    return star_scene, depth_scene


def get_bandwidths(inst, mode, aperture, filter):
    """
    Calculate passband bandwidth properties for photometry modes
    See https://jwst-docs.stsci.edu/jwst-near-infrared-camera/nircam-instrumentation/nircam-filters

    Returns
    -------
    wl0: Float
        Pivot wavelength (Tokunaga & Vacca 2005)
    band_width: Float
        See (Rieke et al. 2008)
    min_wl: Float
        Min wavelvelgth where the response is > 0.25 max(response)
    max_wl: Float
        Max wavelvelgth where the response is > 0.25 max(response)

    Examples
    --------
    >>> import gen_tso.pandeia_io as jwst
    >>> inst = 'miri'
    >>> mode = 'imaging_ts'
    >>> aper = 'imager'
    >>> filter = 'f560w'
    >>> pando = jwst.PandeiaCalculation(inst, mode)
    >>> wl0, bw, min_wl, max_wl = jwst.get_bandwidths(inst, mode, aper, filter)
    """
    throughputs = get_throughputs(inst=inst, mode=mode)
    passband = throughputs[aperture][filter]
    band_wl = passband['wl']
    response = passband['response']

    wl0 = np.sqrt(
        np.trapezoid(response*band_wl, band_wl) /
        np.trapezoid(response/band_wl, band_wl)
    )
    band_width = np.trapezoid(response, band_wl) / np.amax(response)

    response_mask = response > 0.25*np.amax(response)
    min_wl = np.amin(band_wl[response_mask])
    max_wl = np.amax(band_wl[response_mask])
    return wl0, band_width, min_wl, max_wl


def save_tso(filename, tso, lightweight=True):
    """
    Save a TSO output to a pickle file.

    Parameters
    ----------
    filename : String
        The path where the TSO object will be saved.
    tso : dict
        The TSO object to be saved. This should be a dictionary-like
        structure as returned by pando.tso_calculation().
    lightweight : bool
        If True, remove the '2d' and '3d' fields from 'report_out' and
        'report_in' to reduce the file size.
        The original `tso` object is not modified.
    """
    tso_copy = copy.deepcopy(tso)
    if lightweight:
        for rep in ['report_out', 'report_in']:
            report = tso_copy[rep]
            report.pop('2d', None)
            report.pop('3d', None)

    with open(filename, 'wb') as handle:
        pickle.dump(tso_copy, handle, protocol=4)


def simulate_tso(
        tso, n_obs=1, resolution=None, bins=None, noiseless=False,
    ):
    """
    Given a TSO dict from a pandeia TSO run, simulate a transit/eclipse
    spectrum scaling the noise by the number of observations and
    resampling over wavelength.

    Parameters
    ----------
    tso: dict
    n_obs: integer
        Number of transit/eclipse observations
    resolution: float
        If not None, resample the spectrum at the given resolution.
    bins: integer
        If not None, bin the spectrum in between the edges given
        by this array.
    noiseless: Bool
        If True, do not add scatter noise to the spectrum.

    Returns
    -------
    bin_wl: 1D array
        Wavelengths of binned transit/eclipse spectrum.
    bin_spec: 1D array
        Binned simulated transit/eclipse spectrum.
    bin_err: 1D array
        Uncertainties of bin_spec.
    bin_widths: 1D or 2D array
        For spectra, the 1D bin widths of bin_wl.
        For photometry, an array of shape [1,2] with the (lower,upper)
        widths of the passband relative to bin_wl.

    Examples
    --------
    >>> import gen_tso.pandeia_io as jwst
    >>> import matplotlib.pyplot as plt
    >>>
    >>> # A NIRSpec/BOTS simulation
    >>> pando = jwst.PandeiaCalculation('nirspec', 'bots')
    >>> pando.set_scene(
    >>>     sed_type='phoenix', sed_model='k5v',
    >>>     norm_band='2mass,ks', norm_magnitude=8.351,
    >>> )
    >>> disperser='g395h'
    >>> filter='f290lp'
    >>> subarray='sub2048'
    >>> readout='nrsrapid'
    >>> pando.set_config(disperser, filter, subarray, readout)
    >>> ngroup = pando.saturation_fraction(fraction=80.0)
    >>>
    >>> transit_dur = 2.71
    >>> obs_dur = 6.0
    >>> obs_type = 'transit'
    >>> depth_model = np.loadtxt('WASP80b_transit.dat', unpack=True)
    >>>
    >>> tso = pando.tso_calculation(
    >>>     obs_type, transit_dur, obs_dur, depth_model, ngroup,
    >>> )

    >>> bin_wl, bin_spec, bin_err, bin_widths = jwst.simulate_tso(
    >>>    tso, n_obs=1, resolution=300.0, noiseless=False,
    >>> )
    >>>
    >>> plt.figure(10)
    >>> plt.clf()
    >>> plt.plot(depth_model[0], depth_model[1], c='salmon')
    >>> plt.errorbar(bin_wl, bin_spec, bin_err, fmt='o', c='xkcd:blue', ms=4, mfc='w')
    >>> plt.xlim(2.85, 5.2)
    """
    dt_in = tso['time_in']
    dt_out = tso['time_out']
    flux_in = tso['flux_in'] * n_obs
    flux_out = tso['flux_out'] * n_obs
    var_in = tso['var_in'] * n_obs
    var_out = tso['var_out'] * n_obs
    wl = tso['wl']

    # Photometry
    if len(wl) == 1:
        bin_in = flux_in
        bin_out = flux_out
        bin_vin = var_in
        bin_vout = var_out
        # get throughput's bandwidth
        config = tso['report_in']['input']['configuration']['instrument']
        inst = config['instrument']
        mode = config['mode']
        aperture = config['aperture']
        filter = config['filter']
        wl0, bw, min_wl, max_wl = get_bandwidths(inst, mode, aperture, filter)
        bin_wl = wl
        bin_widths = np.array([wl-min_wl, max_wl-wl]).T

    # Spectroscopy binning
    else:
        wl_min = np.amin(wl)
        wl_max = np.amax(wl)
        if resolution is not None:
            bin_edges = constant_resolution_spectrum(wl_min, wl_max, resolution)
            bin_edges = np.append(bin_edges, wl_max)
        elif bins is not None:
            bin_edges = np.copy(bins)
            if bins[0] > wl_min:
                bin_edges = np.append(wl_min, bin_edges)
            else:
                bin_edges[0] = wl_min
            if bins[-1] < wl_max:
                bin_edges = np.append(bin_edges, wl_max)
            else:
                bin_edges[-1] = wl_max
        else:
            bin_edges = 0.5* (wl[1:] + wl[:-1])

        bin_wl = 0.5 * (bin_edges[1:] + bin_edges[:-1])
        bin_widths = bin_wl - bin_edges[:-1]
        nbins = len(bin_wl)
        bin_out = np.zeros(nbins)
        bin_in = np.zeros(nbins)
        bin_vout = np.zeros(nbins)
        bin_vin = np.zeros(nbins)
        for i in range(nbins):
            bin_mask = (wl>=bin_edges[i]) & (wl<bin_edges[i+1])
            bin_out[i] = np.sum(flux_out[bin_mask])
            bin_in[i] = np.sum(flux_in[bin_mask])
            bin_vout[i] = np.sum(var_out[bin_mask])
            bin_vin[i] = np.sum(var_in[bin_mask])

    bin_err = np.sqrt(
        (dt_out/dt_in/bin_out)**2.0 * bin_vin +
        (bin_in*dt_out/dt_in/bin_out**2.0)**2.0 * bin_vout
    )
    # The numpy random system must have its seed reinitialized in
    # each sub-processes to avoid identical 'random' steps.
    # random.randomint is process- and thread-safe.
    np.random.seed(random.randint(0, 100000))
    rand_noise = np.random.normal(0.0, bin_err)
    bin_spec = 1.0 - (bin_in/dt_in) / (bin_out/dt_out)
    if not noiseless:
        bin_spec += rand_noise

    mask = bin_out > 0
    return bin_wl[mask], bin_spec[mask], bin_err[mask], bin_widths[mask]


def _get_tso_wl_range(tso_run):
    """
    Get the wavelength range covered by a TSO calculation

    Parameters
    ----------
    tso_run: Dictionary
        A TSO calculation output as computed by run_pandeia() in the app.
    wl_scale: String
        Wavelength scale: 'linear' or 'log'.

    Returns
    -------
    min_wl: Float
        Shorter-wavelength boundary.
    max_wl: Float
        Longer-wavelength boundary.
    """
    runs = tso_run['tso']
    if not isinstance(runs, list):
        runs = [runs]

    min_wl = np.zeros(len(runs))
    max_wl = np.zeros(len(runs))
    for i,tso in enumerate(runs):
        config = tso['report_in']['input']['configuration']['instrument']
        inst = config['instrument']
        mode = config['mode']
        aper = config['aperture']
        filter = config['filter']
        if mode in _photo_modes:
            wl0, bw, wl_min, wl_max = get_bandwidths(inst, mode, aper, filter)
            dwl_lo = wl0 - wl_min
            dwl_hi = wl_max - wl0
            if inst == 'nircam' and 'w' in filter:
                ndw = 3
            else:
                ndw = 4
            min_wl[i] = wl0 - ndw*dwl_lo
            max_wl[i] = wl0 + ndw*dwl_hi
        else:
            min_wl[i] = np.amin(tso['wl'])
            max_wl[i] = np.amax(tso['wl'])

    min_wl = np.amin(min_wl)
    max_wl = np.amax(max_wl)

    # 5% margin
    d_wl = 0.025 * (max_wl-min_wl)
    min_wl = np.round(min_wl-d_wl, decimals=2)
    max_wl = np.round(max_wl+d_wl, decimals=2)
    return min_wl, max_wl


def _get_tso_depth_range(tso_run, resolution, units):
    """
    Get the transit/eclipse depth range covered by a TSO calculation

    Parameters
    ----------
    tso_run: Dictionary
        A TSO calculation output as computed by run_pandeia() in the app.
    resolution: Float
        Spectral resolution at which to sample the spectrum.
    units: String
        Depth units, select from: 'none', 'percent', 'ppm'.

    Returns
    -------
    min_depth: Float
        Lower depth boundary (~min depth - 3*sigma_depth).
    max_depth: Float
        Higher depth boundary (~max depth + 3*sigma_depth).
    step: Float
        A quarter of the peak-to-peak depth distance.
    """
    runs = tso_run['tso']
    if not isinstance(runs, list):
        runs = [runs]

    min_wl, max_wl = _get_tso_wl_range(tso_run)
    max_depth = []
    min_depth = []
    for tso in runs:
        bin_wl, bin_spec, bin_err, widths = simulate_tso(
            tso, resolution=resolution, noiseless=True,
        )
        err_median = np.median(bin_err)
        d_min1 = np.amin(tso['depth_spectrum'] - 3*err_median)
        d_max1 = np.amax(tso['depth_spectrum'] + 3*err_median)

        mode = tso['report_in']['input']['configuration']['instrument']['mode']
        if mode in _photo_modes:
            input_wl, input_depth = tso['input_depth']
            wl_min = np.amax([min_wl, np.amin(input_wl)])
            wl_max = np.amin([max_wl, np.amax(input_wl)])
            wl = constant_resolution_spectrum(wl_min, wl_max, resolution)
            depth = bin_spectrum(wl, input_wl, input_depth, gaps='interpolate')
            d_min2 = np.amin(depth)
            d_max2 = np.amax(depth)
        else:
            d_min2 = np.inf
            d_max2 = 0.0

        min_depth.append(np.amin([d_min1, d_min2]))
        max_depth.append(np.amax([d_max1, d_max2]))

    min_depth = np.amin(min_depth) / u(units)
    max_depth = np.amax(max_depth) / u(units)
    step = 0.25*(max_depth - min_depth)
    # Only one significant digit on the step size:
    digits = - Decimal(step).adjusted()
    min_depth = np.round(min_depth, decimals=digits)
    max_depth = np.round(max_depth, decimals=digits)
    step = np.round(step, decimals=digits)
    return min_depth, max_depth, step


def _print_pandeia_exposure(
        inst=None, subarray=None, readout=None, ngroup=None, nint=None,
        config=None, format=None,
    ):
    """
    Return a text showing the total exposure time in seconds and hours.
    Either config or the set of inst, subarray, readout, ngroup, and nint
    values must be defined.

    Examples
    --------
    >>> import gen_tso.pandeia_io as jwst
    >>> import numpy as np
    >>>
    >>> wl = np.logspace(0, 2, 1000)
    >>> depth = [wl, np.tile(0.03, len(wl))]
    >>> pando = jwst.PandeiaCalculation('nircam', 'lw_tsgrism')
    >>> pando.set_scene('phoenix', 'k5v', '2mass,ks', 8.351)
    >>> tso = pando.tso_calculation(
    >>>     'transit', transit_dur=2.1, obs_dur=6.0, depth_model=depth,
    >>>     ngroup=190, readout='rapid', filter='f444w',
    >>> )
    >>>
    >>> # Print from config dictionary:
    >>> config = tso['report_out']['input']['configuration']
    >>> print(jwst._print_pandeia_exposure(config=config))
    Exposure time: 13988.25 s (3.89 h)
    >>>
    >>> # Print from direct input values:
    >>> inst = 'nircam'
    >>> subarray = 'subgrism64'
    >>> readout = 'rapid'
    >>> ngroup = 90
    >>> nint = 150
    >>> text = jwst._print_pandeia_exposure(
    >>>     inst, subarray, readout, ngroup, nint,
    >>> )
    >>> print(text)
    Exposure time: 4650.09 s (1.29 h)
    """
    if config is not None:
        inst = config['instrument']['instrument']
        subarray = config['detector']['subarray']
        readout = config['detector']['readout_pattern']
        nint = config['detector']['nint']
        ngroup = config['detector']['ngroup']

    exp_time = exposure_time(inst, subarray, readout, ngroup, nint)
    exposure_hours = exp_time / 3600.0
    exp_text = f'Exposure time: {exp_time:.2f} s ({exposure_hours:.2f} h)'
    return exp_text


def _print_pandeia_saturation(
        inst=None, subarray=None, readout=None, ngroup=None,
        pixel_rate=None, full_well=None, reports=None, format=None,
        req_saturation=80.0,
    ):
    """
    >>> import gen_tso.pandeia_io as jwst
    >>> import numpy as np

    >>> wl = np.logspace(0, 2, 1000)
    >>> depth = [wl, np.tile(0.03, len(wl))]
    >>> pando = jwst.PandeiaCalculation('nircam', 'lw_tsgrism')
    >>> pando.set_scene('phoenix', 'k5v', '2mass,ks', 8.351)
    >>> tso = pando.tso_calculation(
    >>>     'transit', transit_dur=2.1, obs_dur=6.0, depth_model=depth,
    >>>     ngroup=190, readout='rapid', filter='f444w',
    >>> )

    >>> # Directly print from input values:
    >>> pixel_rate, full_well = jwst.extract_flux_rate(tso, get_max=True)
    >>> inst = 'nircam'
    >>> subarray = 'subgrism64'
    >>> readout = 'rapid'
    >>> ngroup = 90
    >>> nint = 150
    >>> text = jwst._print_pandeia_saturation(
    >>>     inst, subarray, readout, ngroup, pixel_rate, full_well,
    >>>     format='html',
    >>> )

    >>> # Print from tso_calculation() or perform_calculation() output:
    >>> text = jwst._print_pandeia_saturation(
    >>>     reports=[tso], format='html',
    >>> )
    """
    # parse reports
    if reports is not None:
        pixel_rate, full_well = extract_flux_rate(reports, get_max=True)
        # This is a TSO dict
        if 'report_in' in reports[0]:
            report = reports[0]['report_in']
        else:
            report = reports[0]
        config = report['input']['configuration']
        inst = config['instrument']['instrument']
        subarray = config['detector']['subarray']
        readout = config['detector']['readout_pattern']
        ngroup = config['detector']['ngroup']

    sat_time = integration_time(inst, subarray, readout, ngroup)
    sat_fraction = 100.0 * pixel_rate * sat_time / full_well
    saturation = format_text(
        f"{sat_fraction:.1f}%",
        np.round(sat_fraction, decimals=1)>np.round(req_saturation, decimals=1),
        sat_fraction>=100,
        format,
    )

    ngroup_req, ngroup_max = groups_below_saturation(
        [req_saturation, 100.0],
        inst, subarray, readout, pixel_rate, full_well,
    )

    ngroup_req = format_text(
        f"{ngroup_req:d}", ngroup_req==2, ngroup_req<2, format,
    )
    ngroup_max = format_text(
        f"{ngroup_max:d}", ngroup_max==2, ngroup_max<2, format,
    )
    sat_text = (
        f'Max fraction of saturation: {saturation}\n'
        f'ngroup below {req_saturation:.0f}% saturation: {ngroup_req}\n'
        f'ngroup below 100% saturation: {ngroup_max}'
    )
    if format == 'html':
        sat_text = sat_text.replace('\n', '<br>')
    return sat_text


def _print_pandeia_stats(inst, mode, report_in, report_out=None, format=None):
    r"""
    Return a text summarizing the SNR, timings, and backround info
    from a perform_calculation() or a tso_calculation() output.

    Parameters
    ----------
    inst: String
        Instrument name.
    mode: String
        Instrument's mode.
    report_in: Dictionary
        A tso_calculation() or pandeia's perform_calculation() output.
    report_out: Dictionary
        A pandeia's perform_calculation() output.
        If not None, assume that the inputs reports are an in-transit
        and out-of-transit pair.
    format: String
        If None format as plain text (e.g., for print() calls)
        If 'rich' format as rich/colorful text (e.g., for FormattedText())
        If 'html' format as HTML text (e.g., for browser applications)

    Examples
    --------
    >>> import gen_tso.pandeia_io as jwst

    >>> inst = 'nircam'
    >>> readout = 'rapid'
    >>> pando = jwst.PandeiaCalculation(inst, 'lw_tsgrism')
    >>> pando.set_scene('phoenix', 'k5v', '2mass,ks', 8.351)
    >>> result = pando.perform_calculation(
    >>>     ngroup=92, nint=683, readout='rapid', filter='f444w',
    >>> )
    >>> report = jwst.print_pandeia_report(inst, mode, result['scalar'])
    >>> print(report)
    """
    if not isinstance(report_in, list):
        report_in = [report_in]
    rate_in = report_in[0]['extracted_flux']
    if report_out == []:
        report_out = None
    if report_out is not None:
        if not isinstance(report_out, list):
           report_out = [report_out]
        rate_out = report_out[0]['extracted_flux']

    # Take report with more flux in it
    if report_out is None or rate_in > rate_out:
        reports = report_in
    else:
        reports = report_out

    snr = ''
    flux = ''
    flux_std = ''
    pixel_rate = ''
    ref_wave = ''
    extract_area = ''
    bkg_area = ''
    bkg_brightness = ''
    sky = ''
    bkg_flux = ''
    bkg_source = ''
    min_snr = 0.0
    if mode == 'target_acq' and inst in ['miri', 'nirspec']:
        min_snr = 20.0
    if mode == 'target_acq' and inst in ['niriss', 'nircam']:
        min_snr = 30.0

    for report in report_in:
        sn = report['sn']
        snr += format_text(f"{sn:9.1f} ", danger=sn<min_snr, format=format)
        flux += f"{report['extracted_flux']:9.1f} "
        flux_std += f"{report['extracted_noise']:9.1f} "
        pixel_rate += f"{report['brightest_pixel']:9.1f} "
        ref_wave += f"{report['reference_wavelength']:6.2f} "
        extract_area += f"{report['extraction_area']:6.2f} "
        if report_in[0]['background_area'] is not None:
            bkg_area += f"{report['background_area']:6.1f} "
            bkg_brightness += f"{report['background']:6.1f} "
            sky += f"{report['background_sky']:6.2f} "
            bkg_flux += f"{report['background_total']:6.2f} "
            bkg_source += f"{report['contamination']:6.2f} "

    background_info = ''
    if report_in[0]['background_area'] is not None:
        background_info = (
            f"Area of background measurement:        {bkg_area} pixels\n"
            f"Background surface brightness:         {bkg_brightness} MJy/sr\n"
            f"Total sky flux in background aperture: {sky} e-/s\n"
            f"Total flux in background aperture:     {bkg_flux} e-/s\n"
            f"Background flux fraction from scene:   {bkg_source.rstrip()}\n"
        )

    cosmic_rays = f"{report_in[0]['cr_ramp_rate']:9.4f}"

    integs = ''
    duty_cycle = ''
    total_time = ''
    exp_time = ''
    dt_exposure = ''
    dt_integ = ''
    dt_fmt = '8.1f' if report_in[0]['total_exposure_time'] > 100 else '8.3f'
    min_duty = 0.0 if mode == 'target_acq' else 0.49
    reports = report_in[0:1]
    if report_out is not None:
        reports.append(report_out[0])
    for report in reports:
        integs += f"{report['total_integrations']:8d} "
        duty = report['duty_cycle']
        duty_cycle += format_text(
            f"{duty:8.2f} ", warning=duty<min_duty, format=format,
        )
        total_time += f"{report['total_exposure_time']:{dt_fmt}} "
        exp_time += f"{report['all_dithers_time']:{dt_fmt}} "
        dt_exposure += f"{report['exposure_time']:{dt_fmt}} "
        dt_integ += f"{report['measurement_time']:{dt_fmt}} "

    if report_out is None:
        tso_header = ''
    else:
        tso_header = f"{'in-transit':>41}  out-transit\n"
    if mode == 'miri_ts' and len(reports)==4:
        channels = "CH1 CH2 CH3 CH4"
        band_header1 = f"{'':31s}{channels.replace(' ', '       ')}\n"
        band_header2 = f"{'':42s}{channels.replace(' ', '    ')}\n"
    else:
        band_header1 = band_header2 = ''

    summary = (
        f"{band_header1}"
        f"Signal-to-noise ratio    {snr.rstrip()}\n"
        f"Extracted flux           {flux} e-/s\n"
        f"Flux standard deviation  {flux_std} e-/s\n"
        f"Brightest pixel rate     {pixel_rate} e-/s\n\n"

        f"{tso_header}"
        f"Integrations:                    {integs.rstrip()}\n"
        f"Duty cycle:                      {duty_cycle.rstrip()}\n"
        f"Total exposure time:             {total_time} s\n"
        # Ignore exp_time since it always matches total_time (nexp=1)
        #f"Single exposure time:            {exp_time} s\n"
        f"First--last dt per exposure:     {dt_exposure} s\n"
        f"Reset--last dt per integration:  {dt_integ} s\n\n"

        f"{band_header2}"
        f"Reference wavelength:                  {ref_wave} microns\n"
        f"Area of extraction aperture:           {extract_area} pixels\n"
        f"{background_info}"
        f"Number of cosmic rays:   {cosmic_rays}  events/pixel/read"
    )

    if format=='html':
        summary = summary.replace('\n', '<br>')
    return summary


def _print_pandeia_report(reports, format=None):
    """
    Get a text summarizing tso_calculation() or perform_calculation() output.
    Similar to the results panel in the ETC.

    Parameters
    ----------
    reports: Dictionary
        A tso_calculation() or a perform_calculation() output.
        Or a list of them.
    format: String
        If None format as plain text (e.g., for print() calls)
        If 'rich' format as rich/colorful text (e.g., for FormattedText())
        If 'html' format as HTML text (e.g., for browser applications)

    Examples
    --------
    >>> import gen_tso.pandeia_io as jwst

    >>> wl = np.logspace(0, 2, 1000)
    >>> depth = [wl, np.tile(0.03, len(wl))]
    >>> pando = jwst.PandeiaCalculation('nircam', 'lw_tsgrism')
    >>> pando.set_scene('phoenix', 'k5v', '2mass,ks', 8.351)
    >>> tso = pando.tso_calculation(
    >>>     'transit', transit_dur=2.1, obs_dur=6.0, depth_model=depth,
    >>>     ngroup=90, readout='rapid', filter='f444w',
    >>> )

    >>> tso_report = jwst._print_pandeia_report(tso, format=None)
    >>> print(tso_report)

    Exposure time: 21545.44 s (5.98 h)
    Max fraction of saturation: 73.7%
    ngroup below 80% saturation: 97
    ngroup below 100% saturation: 122

    Signal-to-noise ratio       3484.8
    Extracted flux              2043.0  e-/s
    Flux standard deviation        0.6  e-/s
    Brightest pixel rate        1354.7  e-/s

                                   in-transit  out-transit
    Integrations:                         243      452
    Duty cycle:                          0.98     0.98
    Total exposure time:               7533.2  14012.3  s
    First--last dt per exposure:       7533.2  14012.3  s
    Reset--last dt per integration:    7366.4  13702.1  s

    Reference wavelength:                    4.36  microns
    Area of extraction aperture:             4.76  pixels
    Area of background measurement:           6.3  pixels
    Background surface brightness:            0.3  MJy/sr
    Total sky flux in background aperture:   4.45  e-/s
    Total flux in background aperture:      64.12  e-/s
    Background flux fraction from scene:     0.93
    Number of cosmic rays:      0.0072  events/pixel/read
    """
    if not isinstance(reports, list):
        reports = [reports]
    # This is a TSO dict
    if 'report_in' in reports[0]:
        report_in = [report['report_in'] for report in reports]
        report_out = [report['report_out'] for report in reports]
    # This is a perform_calculation dict
    else:
        report_in = reports
        report_out = []

    # Put everything into a list to make things easier to handle:
    if not isinstance(report_in, list):
        report_in = [report_in]
    if not isinstance(report_out, list):
        report_out = [report_out]

    # Exposure
    config = report_in[0]['input']['configuration']
    inst = config['instrument']['instrument']
    mode = config['instrument']['mode']
    subarray = config['detector']['subarray']
    readout = config['detector']['readout_pattern']
    ngroup = config['detector']['ngroup']
    nint = config['detector']['nint']
    if report_out != []:
        nint += report_out[0]['input']['configuration']['detector']['nint']
    text_report = _print_pandeia_exposure(inst, subarray, readout, ngroup, nint)

    # Saturation
    reports = report_in + report_out
    pixel_rate, full_well = extract_flux_rate(reports, get_max=True)
    saturation_report = _print_pandeia_saturation(
        inst, subarray, readout, ngroup, pixel_rate, full_well,
        format=format,
    )
    text_report = f'{text_report}\n{saturation_report}'

    # Full report
    scalar_in = [report['scalar'] for report in report_in]
    scalar_out = [report['scalar'] for report in report_out]
    stats = _print_pandeia_stats(inst, mode, scalar_in, scalar_out, format)
    text_report = f'{text_report}\n\n{stats}'
    if format == 'html':
        text_report = text_report.replace('\n', '<br>')
    return text_report


def tso_print(calculation, format='rich'):
    """
    Print to screen a tso_calculation() output or a Pandeia's
    perform_calculation() output.

    Parameters
    ----------
    calculation: Dictionary or list of dictionaries
        A tso_calculation() or a Pandeia's perform_calculation() output.
    format: String
        If 'rich' print with colourful text when there are warnings
        or errors in values.
        If None, print as plain text.

    Examples
    --------
    >>> import gen_tso.pandeia_io as jwst
    >>> import numpy as np

    >>> wl = np.logspace(0, 2, 1000)
    >>> depth = [wl, np.tile(0.03, len(wl))]
    >>> pando = jwst.PandeiaCalculation('nircam', 'lw_tsgrism')
    >>> pando.set_scene('phoenix', 'k5v', '2mass,ks', 8.351)
    >>> tso = pando.tso_calculation(
    >>>     'transit', transit_dur=2.1, obs_dur=6.0, depth_model=depth,
    >>>     ngroup=100, readout='rapid', filter='f444w',
    >>> )
    >>> # A perform_calculation() call:
    >>> jwst.tso_print(tso)

    >>> # A tso_calculation() call:
    >>> calc = pando.perform_calculation(ngroup=130, nint=300)
    >>> jwst.tso_print(calc)
    """
    # TBD: get style from style.css file?
    style = prompt_toolkit.styles.Style.from_dict({
        'danger': '#cb2222',
        'warning': '#ffa500',
    })
    report = _print_pandeia_report(calculation, format)
    if format == 'rich':
        report = prompt_toolkit.HTML(report)
    prompt_toolkit.print_formatted_text(report, style=style)

